@@ -17,6 +17,7 @@
 #include <string.h>
 #include "api.h"
 #include "disjuncts.h"
+#include "disjunct-utils.h"
 #include "error.h"
 #include "preparation.h"
 #include "read-regex.h"
@@ -720,6 +721,8 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	memset(sent, 0, sizeof(struct Sentence_s));
 	sent->dict = dict;
 	sent->length = 0;
+	sent->word = NULL;
+	sent->post_quote = NULL;
 	sent->num_linkages_found = 0;
 	sent->num_linkages_alloced = 0;
 	sent->num_linkages_post_processed = 0;
@@ -747,25 +750,34 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 #ifdef USE_FAT_LINKAGES
 /* XXX Extreme hack alert -- English-language words are used
  * completely naked in the C source code!!! FIXME !!!!
+ * That's OK, this code is deprecated/obsolete, and will be
+ * removed shortly.  We've added the Russian for now, but
+ * really, the Russian dictionary needs to be fixed to support
+ * conjunctions properly.
  */
 static void set_is_conjunction(Sentence sent)
 {
 	int w;
-	char * s;
+	const char * s;
 	for (w=0; w<sent->length; w++) {
-		s = sent->word[w].string;
+		s = sent->word[w].alternatives[0];
 		sent->is_conjunction[w] = 
-			(strcmp(s, "and")==0) ||
-			(strcmp(s, "or" )==0) ||
-			(strcmp(s, "but")==0) ||
-			(strcmp(s, "nor")==0);
+			/* English, obviously ... */
+			(strcmp(s, "and") == 0) ||
+			(strcmp(s, "or" ) == 0) ||
+			(strcmp(s, "but") == 0) ||
+			(strcmp(s, "nor") == 0) ||
+			/* Russian */
+			(strcmp(s, "и") == 0) ||
+			(strcmp(s, "или") == 0) ||
+			(strcmp(s, "но") == 0) ||
+			(strcmp(s, "ни") == 0);
 	}
 }
 #endif /* USE_FAT_LINKAGES */
 
 int sentence_split(Sentence sent, Parse_Options opts)
 {
-	int i;
 	Dictionary dict = sent->dict;
 
 	/* Cleanup stuff previously allocated. This is because some free 
@@ -789,13 +801,6 @@ int sentence_split(Sentence sent, Parse_Options opts)
 	initialize_conjunction_tables(sent);
 #endif /* USE_FAT_LINKAGES */
 
-	for (i=0; i<sent->length; i++)
-	{
-		/* in case we free these before they set to anything else */
-		sent->word[i].x = NULL;
-		sent->word[i].d = NULL;
-	}
-
 	if (!(dict->unknown_word_defined && dict->use_unknown_word))
 	{
 		if (!sentence_in_dictionary(sent)) {
@@ -806,23 +811,33 @@ int sentence_split(Sentence sent, Parse_Options opts)
 	/* Look up each word in the dictionary, collect up all
 	 * plausible disjunct expressions for each word.
 	 */
-	if (!build_sentence_expressions(sent, opts))
-	{
-		sent->num_valid_linkages = 0;
-		return -3;
-	}
+	build_sentence_expressions(sent, opts);
 
 	return 0;
 }
 
+static void free_sentence_words(Sentence sent)
+{
+	int i;
+	for (i = 0; i < sent->length; i++)
+	{
+		free_X_nodes(sent->word[i].x);
+		free_disjuncts(sent->word[i].d);
+		free(sent->word[i].alternatives);
+	}
+	free(sent->word);
+	sent->word = NULL;
+	free(sent->post_quote);
+}
+
 void sentence_delete(Sentence sent)
 {
 	if (!sent) return;
 	sat_sentence_delete(sent);
 #ifdef USE_FAT_LINKAGES
-	free_sentence_disjuncts(sent);
+	free_AND_tables(sent);
 #endif /* USE_FAT_LINKAGES */
-	free_sentence_expressions(sent);
+	free_sentence_words(sent);
 	string_set_delete(sent->string_set);
 	if (sent->parse_info) free_parse_info(sent->parse_info);
 	free_post_processing(sent);
@@ -846,16 +861,27 @@ int sentence_length(Sentence sent)
 	return sent->length;
 }
 
+/**
+ * Deprecated; fails to indicate which alternative was
+ * actually used in the parse!  Well, OK, this works for
+ * ENglish, but is a failure for Russian.
+ */
 const char * sentence_get_word(Sentence sent, int index)
 {
 	if (!sent) return NULL;
-	return sent->word[index].string;
+	if (index >= sent->length) return NULL;
+	return sent->word[index].alternatives[0];
 }
 
+/**
+ * Deprecated; fails to indicate which alternative was
+ * actually used in the parse!
+ */
 const char * sentence_get_nth_word(Sentence sent, int index)
 {
 	if (!sent) return NULL;
-	return sent->word[index].string;
+	if (index >= sent->length) return NULL;
+	return sent->word[index].alternatives[0];
 }
 
 int sentence_null_count(Sentence sent) {
@@ -907,7 +933,8 @@ int sentence_disjunct_cost(Sentence sent, int i) {
 	return sent->link_info[i].disjunct_cost;
 }
 
-int sentence_link_cost(Sentence sent, int i) {
+int sentence_link_cost(Sentence sent, int i)
+{
 	if (!sent) return 0;
 
 	/* The sat solver (currently) fails to fill in link_info */
@@ -978,6 +1005,17 @@ static void chart_parse(Sentence sent, Parse_Options opts)
 	free_fast_matcher(sent);
 }
 
+static void free_sentence_disjuncts(Sentence sent)
+{
+	int i;
+
+	for (i = 0; i < sent->length; ++i)
+	{
+		free_disjuncts(sent->word[i].d);
+		sent->word[i].d = NULL;
+	}
+}
+
 int sentence_parse(Sentence sent, Parse_Options opts)
 {
 	int rc;
@@ -1003,8 +1041,9 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	}
 
 	/* Initialize/free any leftover garbage */
+	free_sentence_disjuncts(sent);  /* Is this really needed ??? */
 #ifdef USE_FAT_LINKAGES
-	free_sentence_disjuncts(sent);
+	if (sentence_contains_conjunction(sent)) free_AND_tables(sent);
 #endif /* USE_FAT_LINKAGES */
 	resources_reset_space(opts->resources);
 
