@@ -1,15 +1,15 @@
-/********************************************************************************/
-/* Copyright (c) 2004                                                           */
-/* Daniel Sleator, David Temperley, and John Lafferty                           */
-/* All rights reserved                                                          */
-/*                                                                              */
-/* Use of the link grammar parsing system is subject to the terms of the        */
-/* license set forth in the LICENSE file included with this software,           */ 
-/* and also available at http://www.link.cs.cmu.edu/link/license.html           */
-/* This license allows free redistribution and use in source and binary         */
-/* forms, with or without modification, subject to certain conditions.          */
-/*                                                                              */
-/********************************************************************************/
+/*************************************************************************/
+/* Copyright (c) 2004                                                    */
+/* Daniel Sleator, David Temperley, and John Lafferty                    */
+/* All rights reserved                                                   */
+/*                                                                       */
+/* Use of the link grammar parsing system is subject to the terms of the */
+/* license set forth in the LICENSE file included with this software,    */ 
+/* and also available at http://www.link.cs.cmu.edu/link/license.html    */
+/* This license allows free redistribution and use in source and binary  */
+/* forms, with or without modification, subject to certain conditions.   */
+/*                                                                       */
+/*************************************************************************/
 /********************************************************************** 
   Calling paradigm:
    . call post_process_open() with the name of a knowledge file. This
@@ -22,8 +22,8 @@
        - Call post_process_close_sentence() 
 ***********************************************************************/
 
-#ifndef _POSTPROCESSH_
-#define _POSTPROCESSH_
+#ifndef _POSTPROCESS_H_
+#define _POSTPROCESS_H_
 
 #define PP_FIRST_PASS  1
 #define PP_SECOND_PASS 2
@@ -39,7 +39,6 @@ PP_node *post_process(Postprocessor * pp, Parse_Options opts,
 int      post_process_match(const char *s, const char *t);  /* utility function */
 
 void          free_d_type(D_type_list * dtl);
-D_type_list * copy_d_type(D_type_list * dtl);
-
+D_type_list * copy_d_type(const D_type_list * dtl);
 
 #endif
