@@ -15,15 +15,14 @@
 #include <stdarg.h>
 #include "api.h"
 #include "corpus/corpus.h"
+#include "print-util.h"
 
 const char * trailer(int mode);
 const char * header(int mode);
 
 static void set_centers(const Linkage linkage, int center[], int print_word_0, int N_words_to_print)
 {
-	mbstate_t mbss;
 	int i, len, tot;
-	memset(&mbss, 0, sizeof(mbss));
 
 	tot = 0;
 	if (print_word_0) i=0; else i=1;
@@ -32,8 +31,9 @@ static void set_centers(const Linkage linkage, int center[], int print_word_0, i
 		/* Centers obtained by counting the characters, 
 		 * not the bytes in the string.
 		 * len = strlen(linkage->word[i]);
+		 * len = mbsrtowcs(NULL, &linkage->word[i], 0, &mbss);
 		 */
-		len = mbsrtowcs(NULL, &linkage->word[i], 0, &mbss);
+		len = utf8_strlen(linkage->word[i]);
 		center[i] = tot + (len/2);
 		tot += len+1;
 	}
@@ -51,17 +51,24 @@ typedef struct
 
 /**
  * prints s then prints the last |t|-|s| characters of t.
- *  if s is longer than t, it truncates s.
+ * if s is longer than t, it truncates s.
+ * multi-byte enabled.
  */
-static void left_append_string(String * string, const char * s, const char * t) {
-	int i, j, k;
-	j = strlen(t);
-	k = strlen(s);
-	for (i=0; i<j; i++) {
-		if (i<k) {
-			append_string(string, "%c", s[i]);
-		} else {
-			append_string(string, "%c", t[i]);
+static void left_append_string(String * string, const char * s, const char * t)
+{
+	size_t i;
+	size_t slen = strlen(s);
+	size_t tlen = strlen(t);
+
+	for (i = 0; i < tlen; )
+	{
+		if (i < slen)
+		{
+			i += append_utf8_char(string, &s[i]);
+		}
+		else
+		{
+			i += append_utf8_char(string, &t[i]);
 		}
 	}
 }
@@ -323,26 +330,30 @@ static char * build_linkage_postscript_string(const Linkage linkage, ps_ctxt_t *
 void compute_chosen_words(Sentence sent, Linkage linkage)
 {
 	int i, l;
-	const char *t;
 	char * s, *u;
 	Parse_info pi = sent->parse_info;
 	const char * chosen_words[MAX_SENTENCE];
 	Parse_Options opts = linkage->opts;
 
 	for (i=0; i<sent->length; i++)
-	{   /* get rid of those ugly ".Ixx" */
-		chosen_words[i] = sent->word[i].string;
-		if (pi->chosen_disjuncts[i] == NULL) {  
+	{
+		const char *t;
+		if (pi->chosen_disjuncts[i] == NULL)
+		{
 			/* no disjunct is used on this word because of null-links */
-			t = chosen_words[i];
+
+			/* Well, alternative[0] is not really right, but what else can we print? */
+			t = sent->word[i].alternatives[0];
 			l = strlen(t) + 2;
 			s = (char *) xalloc(l+1);
 			sprintf(s, "[%s]", t);
 			t = string_set_add(s, sent->string_set);
 			xfree(s, l+1);
-			chosen_words[i] = t;
-		} else if (opts->display_word_subscripts) {
+		}
+		else if (opts->display_word_subscripts)
+		{
 			t = pi->chosen_disjuncts[i]->string;        
+			/* get rid of those ugly ".Ixx" */
 			if (is_idiom_word(t)) {
 				l = strlen(t);
 				s = (char *) xalloc(l+1);
@@ -352,12 +363,20 @@ void compute_chosen_words(Sentence sent, Linkage linkage)
 				*u = '\0';
 				t = string_set_add(s, sent->string_set);
 				xfree(s, l+1);
-				chosen_words[i] = t;
-			} else {
-				chosen_words[i] = t;
 			}
 		}
+		else
+		{
+			/* XXX This is wrong, since it fails to indicate what
+			 * was actally used for the parse, which might not actually 
+			 * be alternative 0.  We should use linkage->->word[i] instead,
+			 * and, umm, manually strip the subscript, or something ...
+			 */
+			t = sent->word[i].alternatives[0];
+		}
+		chosen_words[i] = t;
 	}
+
 	if (sent->dict->left_wall_defined) {
 		chosen_words[0] = LEFT_WALL_DISPLAY;
 	}
@@ -375,12 +394,17 @@ void compute_chosen_words(Sentence sent, Linkage linkage)
 #define MAX_HEIGHT 30
 
 /** 
- * String allocated with exalloc.  
+ * Print the indicated linkage into a utf8-diagram.
+ * Works fine for general utf8 multi-byte sentences.
+ * links and connectors are still mostly assumed to be ASCII, though;
+ * to fix this, grep for strlen in he code below.
+ *
+ * Returned string is allocated with exalloc.  
  * Needs to be freed with linkage_free_diagram()
  */
 static char * linkage_print_diagram_ctxt(const Linkage linkage, ps_ctxt_t *pctx)
 {
-	int i, j, k, cl, cr, row, top_row, width, flag;
+	int i, j, k, cl, cr, row, top_row;
 	const char *s;
 	char *t;
 	int print_word_0 = 0, print_word_N = 0, N_wall_connectors, suppressor_used;
@@ -394,11 +418,12 @@ static char * linkage_print_diagram_ctxt(const Linkage linkage, ps_ctxt_t *pctx)
 	char * gr_string;
 	Dictionary dict = linkage->sent->dict;
 	Parse_Options opts = linkage->opts;
-	int x_screen_width = parse_options_get_screen_width(opts);
+	size_t x_screen_width = parse_options_get_screen_width(opts);
 	int N_words_to_print;
 
 	char picture[MAX_HEIGHT][MAX_LINE];
 	char xpicture[MAX_HEIGHT][MAX_LINE];
+	size_t start[MAX_HEIGHT];
 
 
 	string = string_new();
@@ -590,40 +615,72 @@ static char * linkage_print_diagram_ctxt(const Linkage linkage, ps_ctxt_t *pctx)
 	/* we've built the picture, now print it out */
 	
 	if (print_word_0) i = 0; else i = 1;
-	k = 0;
+
+	/* start locations, for each row.  These may vary, due to different
+	 * utf8 character widths */
+	for (row = top_row; row >= 0; row--)
+		start[row] = 0;
 	pctx->N_rows = 0;
 	pctx->row_starts[pctx->N_rows] = 0;
 	pctx->N_rows++;
-	while(i < N_words_to_print) {
-		append_string(string, "\n");
-		width = 0;
+	while (i < N_words_to_print)
+	{
+		/* Count number of multi-byte chars in the words,
+		 * up to the max screen width. */
+		int uwidth = 0;
 		do {
-			width += strlen(linkage->word[i])+1;
+			uwidth += utf8_strlen(linkage->word[i]) + 1;
 			i++;
-		} while((i<N_words_to_print) &&
-			  (width + ((int)strlen(linkage->word[i]))+1 < x_screen_width));
+		} while ((i < N_words_to_print) &&
+			  (uwidth + utf8_strlen(linkage->word[i]) + 1 < x_screen_width));
+
 		pctx->row_starts[pctx->N_rows] = i - (!print_word_0);    /* PS junk */
-		if (i<N_words_to_print) pctx->N_rows++;     /* same */
-		for (row = top_row; row >= 0; row--) {
-			flag = TRUE;
-			for (j=k;flag&&(j<k+width)&&(xpicture[row][j]!='\0'); j++){
-				flag = flag && (xpicture[row][j] == ' ');
+		if (i < N_words_to_print) pctx->N_rows++;     /* same */
+
+		append_string(string, "\n");
+		for (row = top_row; row >= 0; row--)
+		{
+			/* print each row of the picture */
+			/* 'blank' is used solely to detect blank lines */
+			int mbcnt = 0;
+			Boolean blank = TRUE;
+			k = start[row];
+			for (j = k; (mbcnt < uwidth) && (xpicture[row][j] != '\0'); )
+			{
+				size_t n = utf8_next(&xpicture[row][j]);
+				blank = blank && (xpicture[row][j] == ' ');
+				j += n;
+				mbcnt ++;
 			}
-			if (!flag) {
-				for (j=k;(j<k+width)&&(xpicture[row][j]!='\0'); j++){
-					append_string(string, "%c", xpicture[row][j]);
+         start[row] = j;
+
+			if (!blank)
+			{
+				mbcnt = 0;
+				for (j = k; (mbcnt < uwidth) && (xpicture[row][j] != '\0'); )
+				{
+					/* Copy exactly one multi-byte character to buf */
+					j += append_utf8_char(string, &xpicture[row][j]); 
+					mbcnt ++;
 				}
 				append_string(string, "\n");
 			}
 		}
 		append_string(string, "\n");
-		k += width;
 	}
 	gr_string = string_copy(string);
 	string_delete(string);
 	return gr_string; 
 }
 
+/** 
+ * Print the indicated linkage as utf8-art intp the given string.
+ * The width of the diagram is given by the terminal width, taken
+ * from the parse options.
+ *
+ * The returned string is malloed, and needs to be freed with
+ * linkage_free_diagram()
+ */
 char * linkage_print_diagram(const Linkage linkage)
 {
 	ps_ctxt_t ctx;
@@ -685,7 +742,8 @@ void print_disjunct_counts(Sentence sent)
 		for (d=sent->word[i].d; d != NULL; d = d->next) {
 			c++;
 		}
-		printf("%s(%d) ",sent->word[i].string, c);
+		/* XXX alternatives[0] is not really correct, here .. */
+		printf("%s(%d) ",sent->word[i].alternatives[0], c);
 	}
 	printf("\n\n");
 }
@@ -699,24 +757,12 @@ void print_expression_sizes(Sentence sent)
 		for (x=sent->word[w].x; x!=NULL; x = x->next) {
 			size += size_of_expression(x->exp);
 		}
-		printf("%s[%d] ",sent->word[w].string, size);
+		/* XXX alternatives[0] is not really correct, here .. */
+		printf("%s[%d] ",sent->word[w].alternatives[0], size);
 	}
 	printf("\n\n");
 }
 
-/**
- * this version just prints it on one line. 
- */
-void print_sentence(FILE *fp, Sentence sent, int w)
-{
-	int i;
-	if (sent->dict->left_wall_defined) i=1; else i=0;
-	for (; i<sent->length - sent->dict->right_wall_defined; i++) {
-		fprintf(fp, "%s ", sent->word[i].string);
-	}
-	fprintf(fp, "\n");
-}
-
 const char * trailer(int mode)
 {
 	static const char * trailer_string=
