@@ -18,6 +18,8 @@
 #include "api.h"
 #include "disjunct-utils.h"
 #include "error.h"
+#include "regex-morph.h"
+#include "word-file.h"
 
 const char * linkgrammar_get_version(void)
 {
@@ -130,30 +132,36 @@ const char * linkgrammar_get_dict_version(Dictionary dict)
   generated for idioms.
 
   Dictionary words may be followed by a dot (period, "."), and a "subscript"
-  identifying the word type. The subscript may be one or more letters or 
+  identifying the word type. The subscript may be one or more letters or
   numbers, but must begin with a letter. Currently, the dictionary contains
   (mostly?) subscripts consisting of a single letter, and these serve mostly
   to identify the part-of-speech. In general, subscripts can also be used
   to distinguish different word senses.
 */
 
-static int link_advance(Dictionary dict);
+static Boolean link_advance(Dictionary dict);
 
 static void dict_error2(Dictionary dict, const char * s, const char *s2)
 {
+#define ERRBUFLEN 1024
+	char tokens[ERRBUFLEN], t[ERRBUFLEN];
+	int pos = 1;
 	int i;
-	char tokens[1024], t[128];
 
+	/* The link_advance used to print the error emssage can
+	 * throw more errors... */
 	if (dict->recursive_error) return;
 	dict->recursive_error = TRUE;
 
 	tokens[0] = '\0';
 	for (i=0; i<5 && dict->token[0] != '\0' ; i++)
 	{
-		sprintf(t, "\"%s\" ", dict->token);
-		strcat(tokens, t);
+		pos += snprintf(t, ERRBUFLEN, "\"%s\" ", dict->token);
+		strncat(tokens, t, ERRBUFLEN-1-pos);
 		link_advance(dict);
 	}
+	tokens[pos] = '\0';
+
 	if (s2)
 	{
 		err_ctxt ec;
@@ -192,7 +200,7 @@ static void warning(Dictionary dict, const char * s)
 /**
  * This gets the next character from the input, eliminating comments.
  * If we're in quote mode, it does not consider the % character for
- * comments.
+ * comments.   Note that the returned chacracter is a wide character!
  */
 static wchar_t get_character(Dictionary dict, int quote_mode)
 {
@@ -219,7 +227,7 @@ static wchar_t get_character(Dictionary dict, int quote_mode)
  * Return true if the input wide-character is one of the special
  * characters used to define the syntax of the dictionary.
  */
-static int is_special(wchar_t wc, mbstate_t *ps)
+static Boolean is_special(wchar_t wc, mbstate_t *ps)
 {
 	char buff[MB_LEN_MAX];
 	int nr = wcrtomb(buff, wc, ps);
@@ -228,10 +236,10 @@ static int is_special(wchar_t wc, mbstate_t *ps)
 }
 
 /**
- * This reads the next token from the input into token.
+ * This reads the next token from the input into 'token'.
  * Return 1 if a character was read, else return 0 (and print a warning).
  */
-static int link_advance(Dictionary dict)
+static Boolean link_advance(Dictionary dict)
 {
 	wchar_t c;
 	int nr, i;
@@ -249,7 +257,7 @@ static int link_advance(Dictionary dict)
 			dict->token[1] = '\0';
 		}
 		dict->already_got_it = '\0';
-		return 1;
+		return TRUE;
 	}
 
 	do { c = get_character(dict, FALSE); } while (iswspace(c));
@@ -261,17 +269,17 @@ static int link_advance(Dictionary dict)
 	{
 		if (i > MAX_TOKEN_LENGTH-3) {  /* 3 for multi-byte tokens */
 			dict_error(dict, "Token too long");
-			return 0;
+			return FALSE;
 		}
 		if (quote_mode) {
 			if (c == '\"') {
 				quote_mode = FALSE;
 				dict->token[i] = '\0';
-				return 1;
+				return TRUE;
 			}
 			if (iswspace(c)) {
 				dict_error(dict, "White space inside of token");
-				return 0;
+				return FALSE;
 			}
 
 			/* Although we read wide chars, we store UTF8 internally, always. */
@@ -284,7 +292,7 @@ static int link_advance(Dictionary dict)
 #else
 				dict_error(dict, "Unable to read UTF8 string in current locale");
 #endif
-				return 0;
+				return FALSE;
 			}
 			i += nr;
 		} else {
@@ -295,24 +303,24 @@ static int link_advance(Dictionary dict)
 					dict->token[0] = c;  /* special toks are one char always */
 					dict->token[1] = '\0';
 					dict->is_special = TRUE;
-					return 1;
+					return TRUE;
 				}
 				dict->token[i] = '\0';
 				dict->already_got_it = c;
-				return 1;
+				return TRUE;
 			}
 			if (c == 0x0) {
 				if (i == 0) {
 					dict->token[0] = '\0';
-					return 1;
+					return TRUE;
 				}
 				dict->token[i] = '\0';
 				dict->already_got_it = c;
-				return 1;
+				return TRUE;
 			}
 			if (iswspace(c)) {
 				dict->token[i] = '\0';
-				return 1;
+				return TRUE;
 			}
 			if (c == '\"') {
 				quote_mode = TRUE;
@@ -327,14 +335,14 @@ static int link_advance(Dictionary dict)
 #else
 					dict_error(dict, "Unable to read UTF8 string in current locale");
 #endif
-					return 0;
+					return FALSE;
 				}
 				i += nr;
 			}
 		}
 		c = get_character(dict, quote_mode);
 	}
-	return 1;
+	return TRUE;
 }
 
 /**
@@ -347,7 +355,7 @@ static int is_equal(Dictionary dict, wint_t c)
 	        dict->token[1] == '\0');
 }
 
-/** 
+/**
  * Make sure the string s is a valid connector.
  * Return 1 if the connector is valid, else return 0,
  * and print an appropriate warning message.
@@ -399,16 +407,16 @@ static int check_connector(Dictionary dict, const char * s)
  * example, "make" < "make.n" < "make-up" -- suffixed words come after
  * the bare words, but before any other other words with non-ascii-alpha
  * characters (such as the hyphen in "make-up", or possibly UTF8
- * characters). Thus, stright "strcmp" can't be used to determine 
- * dictionary order. 
+ * characters). Thus, stright "strcmp" can't be used to determine
+ * dictionary order.
  *
  * Thus, a set of specialized string comparison and ordering functions
  * are provided. These "do the right thing" when matching string with
  * and without suffixes.
  */
-/** 
+/**
  * dict_order - order two dictionary words in proper sort order.
- * Return zero if the strings match, else return standard 
+ * Return zero if the strings match, else return standard
  * (locale-dependent) UTF8 sort order.
  * XXX but its not UTF-8 ordered ...! Does this matter ???
  */
@@ -456,9 +464,9 @@ static inline int dict_order(const char *s, const char *t)
  *
  * Note: wild-cards are not used in the English dictionaries;
  * nor are they used in the Persian or Arabic dictionaries...
- * perhaps they are used to handle stemming in some other 
+ * perhaps they are used to handle stemming in some other
  * languages ?????? XXX FIXME, these wildcards might just be
- * old dead code ... 
+ * old dead code ...
  */
 static inline int dict_order_wild(const char * s, const char * t)
 {
@@ -478,11 +486,11 @@ static inline int dict_order_wild(const char * s, const char * t)
 
 /**
  * dict_match --  return true if strings match, else false.
- * A "bare" string (one without a suffix) will match any corresponding 
- * string with a suffix; so, for example, "make" and "make.n" are 
+ * A "bare" string (one without a suffix) will match any corresponding
+ * string with a suffix; so, for example, "make" and "make.n" are
  * a match.  If both strings have suffixes, then the suffixes must match.
  *
- * A subscript is the part that followes the last "." in the word, and 
+ * A subscript is the part that followes the last "." in the word, and
  * that does not begin with a digit.
  */
 static int dict_match(const char * s, const char * t)
@@ -562,7 +570,7 @@ static Dict_node * prune_lookup_list(Dict_node *llist, const char * s)
 void free_lookup_list(Dict_node *llist)
 {
 	Dict_node * n;
-	while(llist != NULL)
+	while (llist != NULL)
 	{
 		n = llist->right;
 		free_dict_node(llist);
@@ -650,14 +658,30 @@ Dict_node * abridged_lookup_list(Dictionary dict, const char *s)
 	return llist;
 }
 
-int boolean_dictionary_lookup(Dictionary dict, const char *s)
+Boolean boolean_dictionary_lookup(Dictionary dict, const char *s)
 {
 	Dict_node *llist = dictionary_lookup_list(dict, s);
-	int boool = (llist != NULL);
+	Boolean boool = (llist != NULL);
 	free_lookup_list(llist);
 	return boool;
 }
 
+/**
+ * Return true if word is in dicationary, or if word is matched by
+ * regex.
+ */
+Boolean find_word_in_dict(Dictionary dict, const char * word)
+{
+	const char * regex_name;
+	if (boolean_dictionary_lookup (dict, word)) return TRUE;
+
+	regex_name = match_regex(dict, word);
+	if (NULL == regex_name) return FALSE;
+
+	return boolean_dictionary_lookup(dict, regex_name);
+}
+
+
 /* ======================================================================== */
 /**
  * Allocate a new Exp node and link it into the exp_list for freeing later.
@@ -724,7 +748,7 @@ static Exp * connector(Dictionary dict)
 		}
 		n = make_unary_node(dict, dn->exp);
 		free_lookup_list(dn_head);
-	} 
+	}
 	else
 	{
 		/* If we are here, token is a connector */
@@ -1051,7 +1075,7 @@ static Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok)
 		n->type = OR_type;
 		n->cost = 0.0f;
 		return n;
-	} 
+	}
 
 	return nl;
 }
@@ -1138,13 +1162,13 @@ static Dict_node *rebalance(Dict_node *root)
 
 /* ======================================================================== */
 /* Implementation of the DSW algo for rebalancing a binary tree.
- * The point is -- after building the dictionary tree, we rebalance it 
- * once at the end. This is a **LOT LOT** quicker than maintaing an 
+ * The point is -- after building the dictionary tree, we rebalance it
+ * once at the end. This is a **LOT LOT** quicker than maintaing an
  * AVL tree along the way (less than quarter-of-a-second vs. about
  * a minute or more!) FWIW, the DSW tree is even more balanced than
  * the AVL tree is (its less deep, more full).
  *
- * The DSW algo, with C++ code, is described in 
+ * The DSW algo, with C++ code, is described in
  *
  * Timothy J. Rolfe, "One-Time Binary Search Tree Balancing:
  * The Day/Stout/Warren (DSW) Algorithm", inroads, Vol. 34, No. 4
@@ -1245,7 +1269,7 @@ Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
 		{
 			n->left = newnode;
 			return n;
-		} 
+		}
 		n->left = insert_dict(dict, n->left, newnode);
 		return n;
 		/* return rebalance(n); Uncomment to get an AVL tree */
@@ -1277,9 +1301,9 @@ Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
  * It inserts the list into the dictionary.
  * It does the middle one first, then the left half, then the right.
  *
- * Note: I think this insert middle, then left, then right, has
- * its origins as a lame attempt to hack around the fact that the 
- * resulting binary tree is rather badly unbalanced. This has been 
+ * Note: I think this "insert middle, then left, then right" algo has
+ * its origins as a lame attempt to hack around the fact that the
+ * resulting binary tree is rather badly unbalanced. This has been
  * fixed by using the DSW rebalancing algo. Now, that would seem
  * to render this crazy bisected-insertion algo obsoloete, but ..
  * oddly enough, it seems to make the DSW balancing go really fast!
@@ -1315,7 +1339,7 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 		ec.sent = NULL;
 		err_msg(&ec, Warn, "Warning: Word \"%s\" found near line %d.\n"
 		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
-		        "\tThis word will be ignored.\n",
+		        "\tThis word will be ignored.",
 		        dn->string, dict->line_number);
 		free_dict_node(dn);
 	}
@@ -1325,7 +1349,7 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 		err_ctxt ec;
 		ec.sent = NULL;
 		err_msg(&ec, Warn, "Warning: The word \"%s\" "
-		          "found near line %d of %s matches the following words:\n",
+		          "found near line %d of %s matches the following words:",
 	             dn->string, dict->line_number, dict->name);
 		for (dnx = dn_head; dnx != NULL; dnx = dnx->right) {
 			fprintf(stderr, "\t%s", dnx->string);
@@ -1351,7 +1375,7 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
  * and is terminated by a semi-colon.
  * Add these words to the dictionary.
  */
-static int read_entry(Dictionary dict)
+static Boolean read_entry(Dictionary dict)
 {
 	Exp *n;
 	int i;
@@ -1366,7 +1390,7 @@ static int read_entry(Dictionary dict)
 		if (dict->is_special)
 		{
 			dict_error(dict, "I expected a word but didn\'t get it.");
-			return 0;
+			goto syntax_error;
 		}
 
 		/* If it's a word-file name */
@@ -1379,21 +1403,84 @@ static int read_entry(Dictionary dict)
 			{
 				err_ctxt ec;
 				ec.sent = NULL;
-				err_msg(&ec, Error, "Error opening word file %s\n", dict->token);
-				return 0;
+				err_msg(&ec, Error, "Error opening word file %s", dict->token);
+				return FALSE;
 			}
 		}
+		else if ((dict->token[0] == '#') && (0 == strcmp(dict->token, "#include")))
+		{
+			Boolean rc;
+			wchar_t* instr;
+			char * dict_name;
+			const char * save_name;
+			Boolean save_is_special;
+			const wchar_t * save_input;
+			const wchar_t * save_pin;
+			wint_t save_already_got_it;
+			int save_line_number;
+
+			if (!link_advance(dict)) goto syntax_error;
+
+			dict_name           = strdup(dict->token);
+			save_name           = dict->name;
+			save_is_special     = dict->is_special;
+			save_input          = dict->input;
+			save_pin            = dict->pin;
+			save_already_got_it = dict->already_got_it;
+			save_line_number    = dict->line_number;
+
+			/* OK, token contains the filename to read ... */
+			instr = get_file_contents(dict_name);
+			if (NULL == instr)
+			{
+				prt_error("Error: Could not open dictionary %s", dict_name);
+				goto syntax_error;
+			}
+			dict->input = instr;
+			dict->pin = dict->input;
+
+			/* The line number and dict name are used for error reporting */
+			dict->line_number = 0;
+			dict->name = dict_name;
+
+			/* Now read the thing in. */
+			rc = read_dictionary(dict);
+
+			dict->name           = save_name;
+			dict->is_special     = save_is_special;
+			dict->input          = save_input;
+			dict->pin            = save_pin;
+			dict->already_got_it = save_already_got_it;
+			dict->line_number    = save_line_number;
+
+			free(instr);
+			free(dict_name);
+			if (!rc) goto syntax_error;
+
+			/* when we return, point to the next entry */
+			if (!link_advance(dict)) goto syntax_error;
+
+			/* If a semicolon follows the include, that's OK... ignore it. */
+			if (';' == dict->token[0])
+			{
+				if (!link_advance(dict)) goto syntax_error;
+			}
+
+			return TRUE;
+		}
 		else
 		{
 			dn_new = dict_node_new();
 			dn_new->left = dn;
+			dn_new->right = NULL;
+			dn_new->exp = NULL;
 			dn = dn_new;
 			dn->file = NULL;
 			dn->string = string_set_add(dict->token, dict->string_set);
 		}
 
 		/* Advance to next entry, unless error */
-		if (0 == link_advance(dict)) goto syntax_error;
+		if (!link_advance(dict)) goto syntax_error;
 	}
 
 	/* pass the : */
@@ -1429,11 +1516,11 @@ static int read_entry(Dictionary dict)
 		i++;
 	}
 	insert_list(dict, dn, i);
-	return 1;
+	return TRUE;
 
 syntax_error:
 	free_lookup_list(dn);
-	return 0;
+	return FALSE;
 }
 
 #if ! defined INFIX_NOTATION
@@ -1504,7 +1591,7 @@ static void print_expression_parens(Exp * n, int need_parens)
 	}
 
 	/* Look for optional, and print only that */
-	el = n->u.l; 
+	el = n->u.l;
 	if (el == NULL)
 	{
 		for (i=0; i<icost; i++) printf("[");
@@ -1590,22 +1677,22 @@ void print_dictionary_data(Dictionary dict)
 	rprint_dictionary_data(dict->root);
 }
 
-int read_dictionary(Dictionary dict)
+Boolean read_dictionary(Dictionary dict)
 {
 	if (!link_advance(dict))
 	{
-		return 0;
+		return FALSE;
 	}
 	while (dict->token[0] != '\0')
 	{
 		if (!read_entry(dict))
 		{
-			return 0;
+			return FALSE;
 		}
 	}
 	dict->root = dsw_tree_to_vine(dict->root);
 	dict->root = dsw_vine_to_tree(dict->root, dict->num_entries);
-	return 1;
+	return TRUE;
 }
 
 /* ======================================================================= */
@@ -1742,30 +1829,88 @@ void free_dictionary(Dictionary dict)
 	free_Exp_list(dict->exp_list);
 }
 
+static size_t altlen(const char **arr)
+{
+	size_t len = 0;
+	if (arr)
+		while (arr[len] != NULL) len++;
+	return len;
+}
+
 /**
  *  dict_display_word_info() - display the information about the given word.
  */
-void dict_display_word_info(Dictionary dict, const char * s)
+static void display_word(Dictionary dict, const char * word,
+                         void (*disp_node)(Dict_node*),
+                         void (*recurse)(Dictionary, const char *))
+     
 {
-	Dict_node *dn, *dn_head;
-	Disjunct * d1, * d2;
-	int len;
-	dn_head = dictionary_lookup_list(dict, s);
-	if (dn_head == NULL)
+	Tokenizer toker;
+	const char * regex_name;
+	Dict_node *dn_head;
+
+	dn_head = dictionary_lookup_list(dict, word);
+	if (dn_head)
 	{
-		printf("	\"%s\" matches nothing in the dictionary.\n", s);
+		disp_node(dn_head);
+		free_lookup_list(dn_head);
 		return;
 	}
-	printf("Matches:\n");
-	for (dn = dn_head; dn != NULL; dn = dn->right)
+
+	/* Recurse, if its a regex match */
+	regex_name = match_regex(dict, word);
+	if (regex_name)
 	{
-		len = 0;
-		d1 = build_disjuncts_for_dict_node(dn);
-		for(d2 = d1 ; d2 != NULL; d2 = d2->next)
+		recurse(dict, regex_name);
+		return;
+	}
+
+	/* If word still wasn't found, try splitting it into 
+	 * prefix-stem-suffix, and print the dict entries for those */
+	toker.pref_alternatives = NULL;
+	toker.stem_alternatives = NULL;
+	toker.suff_alternatives = NULL;
+	toker.string_set = dict->string_set;
+	if (split_word (&toker, dict, word))
+	{
+		size_t i;
+		size_t preflen = altlen(toker.pref_alternatives);
+		size_t stemlen = altlen(toker.stem_alternatives);
+		size_t sufflen = altlen(toker.suff_alternatives);
+
+		if (preflen)
 		{
-			len++;
+			printf("\nPrefix ===================================================\n\n");
+			for (i=0; NULL != toker.pref_alternatives[i]; i++)
+				recurse(dict, toker.pref_alternatives[i]);
 		}
-		free_disjuncts(d1);
+		if (stemlen)
+		{
+			printf("\nStem ===================================================\n\n");
+			for (i=0; NULL != toker.stem_alternatives[i]; i++)
+				recurse(dict, toker.stem_alternatives[i]);
+		}
+		if (sufflen)
+		{
+			printf("\nSuffix ===================================================\n\n");
+			for (i=0; NULL != toker.suff_alternatives[i]; i++)
+				recurse(dict, toker.suff_alternatives[i]);
+		}
+		return;
+	}
+
+	printf("	\"%s\" matches nothing in the dictionary.\n", word);
+}
+
+/**
+ * Display the number of disjuncts associated wth this dict node
+ */
+static void display_counts(Dict_node *dn)
+{
+	printf("Matches:\n");
+	for (; dn != NULL; dn = dn->right)
+	{
+		unsigned int len = count_disjunct_for_dict_node(dn);
 		printf("    ");
 		left_print_string(stdout, dn->string,
 			"                         ");
@@ -1776,25 +1921,23 @@ void dict_display_word_info(Dictionary dict, const char * s)
 		}
 		printf("\n");
 	}
-	free_lookup_list(dn_head);
-	return;
 }
 
 /**
- *  dict_display_word_expr() - display the connector info for a given word.
+ *  dict_display_word_info() - display the information about the given word.
  */
-void dict_display_word_expr(Dictionary dict, const char * s)
+void dict_display_word_info(Dictionary dict, const char * word)
 {
-	Dict_node *dn, *dn_head;
+	display_word(dict, word, display_counts, dict_display_word_info);
+}
 
-	dn_head = dictionary_lookup_list(dict, s);
-	if (dn_head == NULL)
-	{
-		printf("	\"%s\" matches nothing in the dictionary.\n", s);
-		return;
-	}
+/**
+ * Display the number of disjuncts associated wth this dict node
+ */
+static void display_expr(Dict_node *dn)
+{
 	printf("\nExpressions:\n");
-	for (dn = dn_head; dn != NULL; dn = dn->right)
+	for (; dn != NULL; dn = dn->right)
 	{
 		printf("    ");
 		left_print_string(stdout, dn->string,
@@ -1802,6 +1945,12 @@ void dict_display_word_expr(Dictionary dict, const char * s)
 		print_expression(dn->exp);
 		printf("\n\n");
 	}
-	free_lookup_list(dn_head);
-	return;
+}
+
+/**
+ *  dict_display_word_expr() - display the connector info for a given word.
+ */
+void dict_display_word_expr(Dictionary dict, const char * word)
+{
+	display_word(dict, word, display_expr, dict_display_word_expr);
 }
