@@ -109,8 +109,9 @@
                                  should not be longer than this */
 
 /*      Some size definitions.  Reduce these for small machines */
-#define MAX_WORD 60           /* maximum number of chars in a word */
-#define MAX_LINE 1500         /* maximum number of chars in a sentence */
+/* MAX_WORD is large, because unicode entries can use a lot of space */
+#define MAX_WORD 180          /* maximum number of bytes in a word */
+#define MAX_LINE 2500         /* maximum number of chars in a sentence */
 #define MAX_DISJUNCT_COST 10000
 
 /* conditional compiling flags */
@@ -141,6 +142,16 @@ typedef signed __int64 s64; /* signed 64-bit integer, even on 32-bit cpus */
 #define PARSE_NUM_OVERFLOW (((s64)1)<<24)  
 #endif
 
+struct Tokenizer_struct
+{
+	String_set  * string_set;
+	const char ** pref_alternatives;
+	const char ** stem_alternatives;
+	const char ** suff_alternatives;
+};
+
+typedef struct Tokenizer_struct Tokenizer;
+
 struct Connector_struct
 {
     short label;
@@ -201,10 +212,10 @@ struct X_node_struct
 
 struct Word_struct
 {
-    char string[MAX_WORD+1];
+    const char **alternatives;
     X_node * x;      /* sentence starts out with these */
     Disjunct * d;    /* eventually these get generated */
-    int firstupper;
+    Boolean firstupper;
 };
 
 /** 
@@ -345,14 +356,6 @@ struct List_o_links_struct
 #define GENTLE 1
 #define RUTHLESS 0
 
-typedef struct string_node_struct String_node;
-struct string_node_struct
-{
-    char * string;
-    String_node * next;
-    int size;
-};
-
 typedef struct Parse_choice_struct Parse_choice;
 
 struct Link_s
