@@ -1,7 +1,7 @@
 /*************************************************************************/
 /* Copyright (c) 2004                                                    */
 /* Daniel Sleator, David Temperley, and John Lafferty                    */
-/* Copyright (c) 2009 Linas Vepstas                                      */
+/* Copyright (c) 2009, 2012 Linas Vepstas                                */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
@@ -45,14 +45,14 @@
  * names -- e.g. "Sun State Bank" -- "sun", "state" and "bank" are all 
  * common nouns.
  */
-static int is_common_entity(Dictionary dict, const char * str)
+static Boolean is_common_entity(Dictionary dict, const char * str)
 {
 	if (word_contains(dict, str, COMMON_ENTITY_MARKER) == 1)
 		return TRUE;
 	return FALSE;
 }
 
-static int is_entity(Dictionary dict, const char * str)
+static Boolean is_entity(Dictionary dict, const char * str)
 {
 	const char * regex_name;
 	if (word_contains(dict, str, ENTITY_MARKER) == 1)
@@ -73,7 +73,7 @@ static int is_entity(Dictionary dict, const char * str)
  * Basically, if word starts with upper-case latter, we assume
  * its a proper name, and that's that.
  */
-static int is_proper_name(const char * word)
+static Boolean is_proper_name(const char * word)
 {
 	return is_utf8_upper(word);
 }
@@ -104,7 +104,7 @@ static const wchar_t *list_of_quotes(void)
 /**
  * Return TRUE if the character is a quotation character.
  */
-static int is_quote(wchar_t wc)
+static Boolean is_quote(wchar_t wc)
 {
 	static const wchar_t *quotes = NULL;
 	if (NULL == quotes) quotes = list_of_quotes();
@@ -118,7 +118,7 @@ static int is_quote(wchar_t wc)
  * The ":" is included here so we allow "10:30" to be a number.
  * We also allow U+00A0 "no-break space"
  */
-static int is_number(const char * s)
+static Boolean is_number(const char * s)
 {
 	mbstate_t mbs;
 	int nb = 1;
@@ -143,7 +143,7 @@ static int is_number(const char * s)
 /**
  * Returns true if the word contains digits. 
  */
-static int contains_digits(const char * s)
+static Boolean contains_digits(const char * s)
 {
 	mbstate_t mbs;
 	int nb = 1;
@@ -165,39 +165,195 @@ static int contains_digits(const char * s)
  * as a bona-fide word in the sentence.
  *
  * Do not issue the empty string.  
- * Return false if too many words or the word is too long. 
  */
-static int issue_sentence_word(Sentence sent, const char * s)
+static void issue_sentence_word(Sentence sent, const char * s, Boolean quote_found)
 {
-	if (*s == '\0') return TRUE;
-	if (strlen(s) > MAX_WORD)
-	{
-		err_ctxt ec;
-		ec.sent = sent;
-		err_msg(&ec, Error, 
-		   "Error separating sentence. The word \"%s\" is too long.\n"
-		   "A word can have a maximum of %d characters.\n", s, MAX_WORD);
-		return FALSE;
-	}
+	size_t len = sent->length;
+	if (*s == '\0') return;
 
-	if (sent->length >= MAX_SENTENCE)
-	{
-		err_ctxt ec;
-		ec.sent = sent;
-		err_msg(&ec, Error, 
-		   "Error separating sentence. The sentence has too many words.\n");
-		return FALSE;
-	}
+	sent->post_quote = realloc(sent->post_quote, (len+1) * sizeof(Boolean));
+	sent->post_quote[len] = quote_found;
 
-	strcpy(sent->word[sent->length].string, s);
+	sent->word = realloc(sent->word, (len+1) * sizeof(Word));
+	sent->word[len].x = NULL;
+	sent->word[len].d = NULL;
+
+	sent->word[len].alternatives = (const char **) malloc(2*sizeof(const char *));
+	sent->word[len].alternatives[0] = string_set_add(s, sent->string_set);
+	sent->word[len].alternatives[1] = NULL;
 
 	/* Now we record whether the first character of the word is upper-case.
 	   (The first character may be made lower-case
 	   later, but we may want to get at the original version) */
-	if (is_utf8_upper(s)) sent->word[sent->length].firstupper=1;
-	else sent->word[sent->length].firstupper = 0;
+	if (is_utf8_upper(s)) sent->word[len].firstupper = TRUE;
+	else sent->word[len].firstupper = FALSE;
+
 	sent->length++;
-	return TRUE;
+}
+
+static size_t altlen(const char **arr)
+{
+	size_t len = 0;
+	if (arr)
+		while (arr[len] != NULL) len++;
+	return len;
+}
+
+static const char ** resize_alts(const char **arr, size_t len)
+{
+	arr = realloc(arr, (len+2) * sizeof(const char *));
+	arr[len+1] = NULL;
+	return arr;
+}
+
+/**
+ * Accumulate different word-stemming possibilities 
+ */
+static void add_suffix_alternatives(Tokenizer * tok,
+                                   const char * stem, const char * suffix)
+{
+	char buff[MAX_WORD+1];
+	size_t sz;
+	
+	size_t stemlen = altlen(tok->stem_alternatives);
+	size_t sufflen = altlen(tok->suff_alternatives);
+
+	tok->stem_alternatives = resize_alts(tok->stem_alternatives, stemlen);
+	tok->stem_alternatives[stemlen] = string_set_add(stem, tok->string_set);
+
+	if (NULL == suffix)
+		return;
+
+	tok->suff_alternatives = resize_alts(tok->suff_alternatives, sufflen);
+
+	/* Add an equals-sign to the suffix.  This is needed to distinguish
+	 * suffixes that were stripped off from ordinary words that just
+	 * happen to be the same as the suffix. Kind-of a weird hack,
+	 * but I'm not sure what else to do... */
+	sz = MIN(strlen(suffix) + 2, MAX_WORD);
+	buff[0] = '=';
+	strncpy(&buff[1], suffix, sz);
+	buff[sz] = 0;
+
+	tok->suff_alternatives[sufflen] = string_set_add(buff, tok->string_set);
+}
+
+static void add_presuff_alternatives(Tokenizer * tok, const char * prefix,
+                                    const char * stem, const char * suffix)
+{
+	char buff[MAX_WORD+1];
+	size_t sz;
+
+	size_t preflen = altlen(tok->pref_alternatives);
+	size_t stemlen = altlen(tok->stem_alternatives);
+	size_t sufflen = altlen(tok->suff_alternatives);
+
+	tok->pref_alternatives = resize_alts(tok->pref_alternatives, preflen);
+	tok->pref_alternatives[preflen] = string_set_add(prefix, tok->string_set);
+
+	tok->stem_alternatives = resize_alts(tok->stem_alternatives, stemlen);
+	tok->stem_alternatives[stemlen] = string_set_add(stem, tok->string_set);
+
+	if (NULL == suffix)
+		return;
+
+	tok->suff_alternatives = resize_alts(tok->suff_alternatives, sufflen);
+
+	/* Add an equals-sign to the suffix.  This is needed to distinguish
+	 * suffixes that were stripped off from ordinary words that just
+	 * happen to be the same as the suffix. Kind-of a weird hack,
+	 * but I'm not sure what else to do... */
+	sz = MIN(strlen(suffix) + 2, MAX_WORD);
+	buff[0] = '=';
+	strncpy(&buff[1], suffix, sz);
+	buff[sz] = 0;
+
+	tok->suff_alternatives[sufflen] = string_set_add(buff, tok->string_set);
+}
+
+/**
+ * Same as issue_sentence_word, except that here, we issue multiple
+ * alternative prefix-stem-suffix possibilities. Up to three "words"
+ * may be issued.
+ */
+static Boolean issue_alternatives(Sentence sent, Boolean quote_found)
+{
+	Boolean issued = FALSE;
+	Tokenizer *tokenizer = &sent->tokenizer;
+	size_t len = sent->length;
+	size_t preflen = altlen(tokenizer->pref_alternatives);
+	size_t stemlen = altlen(tokenizer->stem_alternatives);
+	size_t sufflen = altlen(tokenizer->suff_alternatives);
+
+	if (preflen)
+	{
+		size_t i;
+
+		sent->post_quote = realloc(sent->post_quote, (len+1) * sizeof(Boolean));
+		sent->post_quote[len] = quote_found;
+
+		sent->word = realloc(sent->word, (len+1) * sizeof(Word));
+		sent->word[len].x = NULL;
+		sent->word[len].d = NULL;
+
+		sent->word[len].alternatives = tokenizer->pref_alternatives;
+		tokenizer->pref_alternatives = NULL;
+
+		sent->word[len].firstupper = FALSE;
+		for (i=0; NULL != tokenizer->pref_alternatives[i]; i++) {
+			const char *s = tokenizer->pref_alternatives[i];
+		   if (is_utf8_upper(s)) sent->word[len].firstupper = TRUE;
+		}
+
+		len++;
+		issued = TRUE;
+	}
+
+	if (stemlen)
+	{
+		sent->post_quote = realloc(sent->post_quote, (len+1) * sizeof(Boolean));
+		sent->post_quote[len] = preflen ? FALSE : quote_found;
+
+		sent->word = realloc(sent->word, (len+1) * sizeof(Word));
+		sent->word[len].x = NULL;
+		sent->word[len].d = NULL;
+
+		sent->word[len].alternatives = tokenizer->stem_alternatives;
+
+		sent->word[len].firstupper = FALSE;
+		if (0 == preflen)
+		{
+			size_t i;
+			for (i = 0; NULL != tokenizer->stem_alternatives[i]; i++) {
+				const char *s = tokenizer->stem_alternatives[i];
+		   	if (is_utf8_upper(s)) sent->word[len].firstupper = TRUE;
+			}
+		}
+
+		tokenizer->stem_alternatives = NULL;
+		len++;
+		issued = TRUE;
+	}
+
+	if (sufflen)
+	{
+		sent->post_quote = realloc(sent->post_quote, (len+1) * sizeof(Boolean));
+		sent->post_quote[len] = FALSE;
+
+		sent->word = realloc(sent->word, (len+1) * sizeof(Word));
+		sent->word[len].x = NULL;
+		sent->word[len].d = NULL;
+
+		sent->word[len].alternatives = tokenizer->suff_alternatives;
+		tokenizer->suff_alternatives = NULL;
+
+		sent->word[len].firstupper = FALSE;
+		len++;
+		issued = TRUE;
+	}
+
+	sent->length = len;
+	return issued;
 }
 
 /*
@@ -255,48 +411,114 @@ static int issue_sentence_word(Sentence sent, const char * s)
 #undef		MIN
 #define MIN(a, b)  (((a) < (b)) ? (a) : (b))
 
-static int boolean_reg_dict_lookup(Dictionary dict, const char * word)
+/**
+ * Split word into prefix, stem and suffix. 
+ * Record all of the alternative splittings in the tokenizer
+ * alternatives arrays.
+ */
+static Boolean suffix_split(Tokenizer *tokenizer, Dictionary dict,
+                            const char *w, const char *wend)
 {
-	const char * regex_name;
-	if (boolean_dictionary_lookup(dict, word)) return TRUE;
+	int i, j, len;
+	int p_strippable, s_strippable;
+	const char **prefix, **suffix;
+	char newword[MAX_WORD+1];
+	Boolean word_is_in_dict = FALSE;
 
-	regex_name = match_regex(dict, word);
-	if (NULL == regex_name) return FALSE;
+	/* Set up affix tables.  */
+	Dictionary afdict = dict->affix_table;
+	if (NULL == afdict) return FALSE;
 
-	return boolean_dictionary_lookup(dict, regex_name);
-}
+	p_strippable = afdict->p_strippable;
+	s_strippable = afdict->s_strippable;
+	prefix = afdict->prefix;
+	suffix = afdict->suffix;
 
-static int downcase_is_in_dict(Dictionary dict, char * word)
-{
-	int i, rc;
-	char low[MB_LEN_MAX];
-	char save[MB_LEN_MAX];
-	wchar_t c;
-	int nbl, nbh;
-	mbstate_t mbs, mbss;
-
-	if (!is_utf8_upper(word)) return FALSE;
-
-	memset(&mbs, 0, sizeof(mbs));
-	memset(&mbss, 0, sizeof(mbss));
-
-	nbh = mbrtowc (&c, word, MB_CUR_MAX, &mbs);
-	c = towlower(c);
-	nbl = wctomb_check(low, c, &mbss);
-	if (nbh != nbl)
+	j = 0;
+	for (i=0; i <= s_strippable; i++)
 	{
-		prt_error("Warning: can't downcase multi-byte string: %s\n", word);
-		return FALSE;
+		Boolean s_ok = FALSE;
+		/* Go through once for each suffix; then go through one 
+		 * final time for the no-suffix case (i.e. to look for
+		 * prefixes only, without suffixes.) */
+		if (i < s_strippable)
+		{
+			len = strlen(suffix[i]);
+
+			/* The remaining w is too short for a possible match */
+			if ((wend-w) < len) continue;
+
+			/* Always match the zero-lenght suffix */
+			if (0 == len) s_ok = TRUE;
+			else if (strncmp(wend-len, suffix[i], len) == 0) s_ok = TRUE;
+		}
+		else
+			len = 0;
+
+		if (s_ok || i == s_strippable)
+		{
+			size_t sz = MIN((wend-len)-w, MAX_WORD);
+			strncpy(newword, w, sz);
+			newword[sz] = '\0';
+
+			/* Check if the remainder is in the dictionary;
+			 * for the no-suffix case, it won't be */
+			if ((i < s_strippable) &&
+			    find_word_in_dict(dict, newword))
+			{
+				if (1 < verbosity)
+					printf("Splitting word into two: %s-%s\n", newword, suffix[i]);
+
+				add_suffix_alternatives(tokenizer, newword, suffix[i]);
+				word_is_in_dict = TRUE;
+			}
+
+			/* If the remainder isn't in the dictionary, 
+			 * try stripping off prefixes */
+			else
+			{
+				for (j=0; j<p_strippable; j++)
+				{
+					if (strncmp(w, prefix[j], strlen(prefix[j])) == 0)
+					{
+						int sz = MIN((wend - len) - (w + strlen(prefix[j])), MAX_WORD);
+						strncpy(newword, w+strlen(prefix[j]), sz);
+						newword[sz] = '\0';
+						if (find_word_in_dict(dict, newword))
+						{
+							if (verbosity>1)
+							{
+								if (i < s_strippable)
+									printf("Splitting word into three: %s-%s-%s\n",
+										prefix[j], newword, suffix[i]);
+								else
+									printf("Splitting off a prefix: %s-%s\n",
+										prefix[j], newword);
+							}
+							word_is_in_dict = TRUE;
+							if (i < s_strippable)
+								add_presuff_alternatives(tokenizer, prefix[j], newword, suffix[i]);
+							else
+								add_presuff_alternatives(tokenizer, prefix[j], newword, NULL);
+						}
+					}
+				}
+			}
+			if (j != p_strippable) break;
+		}
 	}
 
-	/* Downcase */
-	for (i=0; i<nbl; i++) { save[i] = word[i]; word[i] = low[i]; }
+	return word_is_in_dict;
+}
 
-	/* Look it up, then restore old value */
-	rc = boolean_reg_dict_lookup(dict, word);
-	for (i=0; i<nbh; i++) { word[i] = save[i]; }
-
-	return rc; 
+/**
+ * Split word into prefix, stem and suffix. 
+ * Record all of the alternative splittings in the tokenizer
+ * alternatives arrays.
+ */
+Boolean split_word(Tokenizer *tokenizer, Dictionary dict, const char *word)
+{
+	return suffix_split(tokenizer, dict, word, word + strlen(word));
 }
 
 /**
@@ -306,17 +528,17 @@ static int downcase_is_in_dict(Dictionary dict, char * word)
  * parts.  The process is described above.  Returns TRUE if OK, FALSE if
  * too many punctuation marks or other separation error.
  */
-static int separate_word(Sentence sent, Parse_Options opts,
+static void separate_word(Sentence sent, Parse_Options opts,
                          const char *w, const char *wend,
-                         int is_first_word, int quote_found)
+                         Boolean quote_found)
 {
 	size_t sz;
-	int i, j, len;
+	int i, len;
 	int r_strippable=0, l_strippable=0, u_strippable=0;
-	int s_strippable=0, p_strippable=0;
-	int  n_r_stripped, s_stripped;
-	int word_is_in_dict, s_ok;
-	int issued = FALSE;
+	int  n_r_stripped = 0;
+	Boolean word_is_in_dict;
+	Boolean issued = FALSE;
+	Boolean have_empty_suffix = FALSE;
 
 	int found_number = 0;
 	int n_r_stripped_save;
@@ -325,48 +547,55 @@ static int separate_word(Sentence sent, Parse_Options opts,
 	const char ** strip_left = NULL;
 	const char ** strip_right = NULL;
 	const char ** strip_units = NULL;
-	const char ** prefix = NULL;
-	const char ** suffix = NULL;
 	char word[MAX_WORD+1];
-	char newword[MAX_WORD+1];
 
 	const char *r_stripped[MAX_STRIP];  /* these were stripped from the right */
 
+	Dictionary dict = sent->dict;
+
+	if (dict->affix_table != NULL)
+	{
+		have_empty_suffix = dict->affix_table->have_empty_suffix;
+	}
+
 	/* First, see if we can already recognize the word as-is. If
 	 * so, then we are done. Else we'll try stripping prefixes, suffixes.
 	 */
 	sz = MIN(wend-w, MAX_WORD);
 	strncpy(word, w, sz);
 	word[sz] = '\0';
-	word_is_in_dict = FALSE;
 
-	if (boolean_reg_dict_lookup(sent->dict, word))
-		word_is_in_dict = TRUE;
-	else if (is_first_word && downcase_is_in_dict (sent->dict,word))
-		word_is_in_dict = TRUE;
+	/* The simplest case: no affixes, suffixes, etc. */
+	word_is_in_dict = find_word_in_dict (dict, word);
 
-	if (word_is_in_dict)
+	/* ... unless its a lang like Russian, which allows empty
+	 * suffixes, which have a real morphological linkage.
+	 * In which case, we have to try them out. */
+	if (word_is_in_dict & !have_empty_suffix)
 	{
-		return issue_sentence_word(sent, word);
+		issue_sentence_word(sent, word, quote_found);
+		return;
 	}
 
 	/* Set up affix tables.  */
-	if (sent->dict->affix_table != NULL)
+	if (dict->affix_table != NULL)
 	{
-		Dictionary dict = sent->dict->affix_table;
-		r_strippable = dict->r_strippable;
-		l_strippable = dict->l_strippable;
-		u_strippable = dict->u_strippable;
-		p_strippable = dict->p_strippable;
-		s_strippable = dict->s_strippable;
+		Dictionary afdict = dict->affix_table;
+		r_strippable = afdict->r_strippable;
+		l_strippable = afdict->l_strippable;
+		u_strippable = afdict->u_strippable;
 
-		strip_left = dict->strip_left;
-		strip_right = dict->strip_right;
-		strip_units = dict->strip_units;
-		prefix = dict->prefix;
-		suffix = dict->suffix;
+		strip_left = afdict->strip_left;
+		strip_right = afdict->strip_right;
+		strip_units = afdict->strip_units;
 	}
 
+	/* If we found the word in the dict, but it also can have an
+	 * empty suffix, then just skip right over to suffix processing
+	 */
+	if (word_is_in_dict && have_empty_suffix)
+		goto do_suffix_processing;
+
 	/* Strip off punctuation, etc. on the left-hand side. */
 	/* XXX FIXME: this fails in certain cases: e.g.
 	 * "By the '50s, he was very prosperous."
@@ -374,8 +603,9 @@ static int separate_word(Sentence sent, Parse_Options opts,
 	 * found in the dict.  Next, the comma is removed, and "50s" is still
 	 * not in the dict ... the trick was that the comma should be 
 	 * right-stripped first, then the possible quotes. 
-	 * More generally, link-grammar does not support multiple possible
-	 * tokenizations.
+	 * More generally, the current implementation of the link-parser algo
+	 * does not support multiple alternative tokenizations; the viterbi
+	 * parser, under development, should be able to do better.
 	 */
 	for (;;)
 	{
@@ -385,7 +615,7 @@ static int separate_word(Sentence sent, Parse_Options opts,
 			sz = strlen(strip_left[i]);
 			if (strncmp(w, strip_left[i], sz) == 0)
 			{
-				if (!issue_sentence_word(sent, strip_left[i])) return FALSE;
+				issue_sentence_word(sent, strip_left[i], quote_found);
 				w += sz;
 				break;
 			}
@@ -397,7 +627,7 @@ static int separate_word(Sentence sent, Parse_Options opts,
 	 * left-punctuation, in which case, it has all been issued.
 	 * So -- we're done, return.
 	 */
-	if (w >= wend) return TRUE;
+	if (w >= wend) return;
 
 	/* Now w points to the string starting just to the right of
 	 * any left-stripped characters.
@@ -414,14 +644,7 @@ static int separate_word(Sentence sent, Parse_Options opts,
 		word[sz] = '\0';
 		if (wend == w) break;  /* it will work without this */
 
-		if (boolean_reg_dict_lookup(sent->dict, word))
-		{
-			word_is_in_dict = TRUE;
-			break;
-		}
-
-		/* This could happen if it's a word after a colon, also! */
-		if (is_first_word && downcase_is_in_dict (sent->dict, word))
+		if (find_word_in_dict(dict, word))
 		{
 			word_is_in_dict = TRUE;
 			break;
@@ -494,92 +717,30 @@ static int separate_word(Sentence sent, Parse_Options opts,
 		}
 	}
 
-	/* Now we strip off suffixes...w points to the remaining word, 
-	 * "wend" to the end of the word. */
-
-	s_stripped = -1;
-	strncpy(word, w, MIN(wend-w, MAX_WORD));
-	word[MIN(wend-w, MAX_WORD)] = '\0';
-
-	/* Umm, double-check, if need be ... !?? */
+	/* Umm, double-check, if need be ... !??  Why? */
 	if (FALSE == word_is_in_dict)
 	{
-		if (boolean_reg_dict_lookup(sent->dict, word))
-			word_is_in_dict = TRUE;
-		else if (is_first_word && downcase_is_in_dict (sent->dict,word))
-			word_is_in_dict = TRUE;
+		/* w points to the remaining word, 
+i		 * "wend" to the end of the word. */
+		sz = MIN(wend-w, MAX_WORD);
+		strncpy(word, w, sz);
+		word[sz] = '\0';
+
+		word_is_in_dict = find_word_in_dict(dict, word);
 	}
 
-	if (FALSE == word_is_in_dict)
+do_suffix_processing:
+	/* OK, now try to strip suffixes. */
+	if (!word_is_in_dict || have_empty_suffix)
 	{
-		j=0;
-		for (i=0; i <= s_strippable; i++)
-		{
-			s_ok = 0;
-			/* Go through once for each suffix; then go through one 
-			 * final time for the no-suffix case */
-			if (i < s_strippable)
-			{
-				len = strlen(suffix[i]);
+		Tokenizer *tokenizer = &sent->tokenizer;
 
-				/* The remaining w is too short for a possible match */
-				if ((wend-w) < len) continue;
-				if (strncmp(wend-len, suffix[i], len) == 0) s_ok=1;
-			}
-			else
-				len = 0;
+		tokenizer->string_set = dict->string_set;
+		tokenizer->pref_alternatives = NULL;
+		tokenizer->stem_alternatives = NULL;
+		tokenizer->suff_alternatives = NULL;
 
-			if (s_ok || i == s_strippable)
-			{
-				strncpy(newword, w, MIN((wend-len)-w, MAX_WORD));
-				newword[MIN((wend-len)-w, MAX_WORD)] = '\0';
-
-				/* Check if the remainder is in the dictionary;
-				 * for the no-suffix case, it won't be */
-				if (boolean_reg_dict_lookup(sent->dict, newword))
-				{
-					if ((verbosity>1) && (i < s_strippable))
-						printf("Splitting word into two: %s-%s\n", newword, suffix[i]);
-					s_stripped = i;
-					wend -= len;
-					strncpy(word, w, MIN(wend-w, MAX_WORD));
-					word[MIN(wend-w, MAX_WORD)] = '\0';
-					word_is_in_dict = TRUE;
-					break;
-				}
-
-				/* If the remainder isn't in the dictionary, 
-				 * try stripping off prefixes */
-				else
-				{
-					for (j=0; j<p_strippable; j++)
-					{
-						if (strncmp(w, prefix[j], strlen(prefix[j])) == 0)
-						{
-							int sz = MIN((wend-len)-(w+strlen(prefix[j])), MAX_WORD);
-							strncpy(newword, w+strlen(prefix[j]), sz);
-							newword[sz] = '\0';
-							if (boolean_reg_dict_lookup(sent->dict, newword))
-							{
-								if ((verbosity>1) && (i < s_strippable))
-									printf("Splitting word into three: %s-%s-%s\n",
-										prefix[j], newword, suffix[i]);
-								if (!issue_sentence_word(sent, prefix[j])) return FALSE;
-								if (i < s_strippable) s_stripped = i;
-								wend -= len;
-								w += strlen(prefix[j]);
-								sz = MIN(wend-w, MAX_WORD);
-								strncpy(word, w, sz);
-								word[sz] = '\0';
-								word_is_in_dict = TRUE;
-								break;
-							}
-						}
-					}
-				}
-				if (j != p_strippable) break;
-			}
-		}
+		word_is_in_dict = suffix_split(tokenizer, dict, w, wend);
 	}
 
 	/* word is now what remains after all the stripping has been done */
@@ -595,8 +756,6 @@ static int separate_word(Sentence sent, Parse_Options opts,
 		word_is_in_dict = TRUE;
 	}
 
-	if (quote_found == TRUE) sent->post_quote[sent->length] = 1;
-
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 	/* If the word is still not being found, then it might be 
 	 * a run-on of two words. Ask the spell-checker to split
@@ -605,33 +764,34 @@ static int separate_word(Sentence sent, Parse_Options opts,
 	 */
 	if ((FALSE == word_is_in_dict) && 
 	    TRUE == opts->use_spell_guess &&
-	    sent->dict->spell_checker &&
+	    dict->spell_checker &&
 	    (FALSE == is_proper_name(word)))
 	{
 		char **alternates = NULL;
 		char *sp = NULL;
 		char *wp;
 		int j, n;
-		n = spellcheck_suggest(sent->dict->spell_checker, &alternates, word);
+		n = spellcheck_suggest(dict->spell_checker, &alternates, word);
 		for (j=0; j<n; j++)
 		{
-			/* Uhh, XXX this is not utf8 safe! */
+			/* Uhh, XXX this is not utf8 safe! ?? */
 			sp = strchr(alternates[j], ' ');
 			if (sp) break;
 		}
 
 		if (sp) issued = TRUE;
 
+		/* It may be three run-on words or more. Loop over all */
 		wp = alternates[j];
 		while (sp)
 		{
 			*sp = 0x0;
-			if (!issue_sentence_word(sent, wp)) return FALSE;
+			issue_sentence_word(sent, wp, quote_found);
 			wp = sp+1;
 			sp = strchr(wp, ' ');
 			if (NULL == sp)
 			{
-				if (!issue_sentence_word(sent, wp)) return FALSE;
+				issue_sentence_word(sent, wp, quote_found);
 			}
 		}
 		if (alternates) spellcheck_free_suggest(alternates, n);
@@ -639,21 +799,15 @@ static int separate_word(Sentence sent, Parse_Options opts,
 #endif /* HAVE_HUNSPELL */
 
 	if (FALSE == issued)
-	{
-		if (!issue_sentence_word(sent, word)) return FALSE;
-	}
+		issued = issue_alternatives(sent, quote_found);
 
-	if (s_stripped != -1)
-	{
-	  if (!issue_sentence_word(sent, suffix[s_stripped])) return FALSE;
-	}
+	if (FALSE == issued)
+		issue_sentence_word(sent, word, quote_found);
 
-	for (i = n_r_stripped-1; i>=0; i--)
+	for (i = n_r_stripped - 1; i >= 0; i--)
 	{
-		if (!issue_sentence_word(sent, r_stripped[i])) return FALSE;
+		issue_sentence_word(sent, r_stripped[i], FALSE);
 	}
-
-	return TRUE;
 }
 
 /**
@@ -662,24 +816,22 @@ static int separate_word(Sentence sent, Parse_Options opts,
  * the sent->word[] array.  Returns TRUE if all is well, FALSE otherwise.
  * Quote marks are treated just like blanks.
  */
-int separate_sentence(Sentence sent, Parse_Options opts)
+Boolean separate_sentence(Sentence sent, Parse_Options opts)
 {
 	const char *t;
-	int is_first, quote_found;
+	Boolean quote_found;
 	Dictionary dict = sent->dict;
 	mbstate_t mbs;
 	const char * s = sent->orig_sentence;
 
-	memset(sent->post_quote, 0, MAX_SENTENCE*sizeof(int));
 	sent->length = 0;
 
 	if (dict->left_wall_defined)
-		if (!issue_sentence_word(sent, LEFT_WALL_WORD)) return FALSE;
+		issue_sentence_word(sent, LEFT_WALL_WORD, FALSE);
 
 	/* Reset the multibyte shift state to the initial state */
 	memset(&mbs, 0, sizeof(mbs));
 
-	is_first = TRUE;
 	for(;;) 
 	{
 		int isq;
@@ -718,16 +870,16 @@ int separate_sentence(Sentence sent, Parse_Options opts)
 			if (0 > nb) goto failure;
 		}
 
-		if (!separate_word(sent, opts, s, t, is_first, quote_found)) return FALSE;
-		is_first = FALSE;
+		separate_word(sent, opts, s, t, quote_found);
 		s = t;
 		if (*s == '\0') break;
 	}
 
 	if (dict->right_wall_defined)
-		if (!issue_sentence_word(sent, RIGHT_WALL_WORD)) return FALSE;
+		issue_sentence_word(sent, RIGHT_WALL_WORD, FALSE);
 
-	return (sent->length > dict->left_wall_defined + dict->right_wall_defined);
+	return (sent->length > (dict->left_wall_defined ? 1 : 0) +
+	       (dict->right_wall_defined ? 1 : 0));
 
 failure:
 	prt_error("Unable to process UTF8 input string in current locale %s\n",
@@ -741,26 +893,28 @@ failure:
  * Also, add a subscript to the resulting word to indicate the
  * rule origin.
  */
-static void tag_regex_string(Sentence sent, int i, const char * type)
+static X_node * build_regex_expressions(Sentence sent, int i,
+                             const char * word_type, const char * word)
 {
-	char str[MAX_WORD+1];
-	char * t;
-	X_node * e;
-	sent->word[i].x = build_word_expressions(sent->dict, type);
-	for (e = sent->word[i].x; e != NULL; e = e->next)
+	X_node *e, *we;
+
+	we = build_word_expressions(sent->dict, word_type);
+	for (e = we; e != NULL; e = e->next)
 	{
+		char str[MAX_WORD+1];
+		char * t;
 		t = strchr(e->string, '.');
-		e->string = sent->word[i].string;
 		if (NULL != t)
 		{
-			snprintf(str, MAX_WORD, "%.50s[!].%.5s", e->string, t+1);
+			snprintf(str, MAX_WORD, "%.50s[!].%.5s", word, t+1);
 		}
 		else
 		{
-			snprintf(str, MAX_WORD, "%.50s", e->string);
+			snprintf(str, MAX_WORD, "%.50s[!]", word);
 		}
 		e->string = string_set_add(str, sent->string_set);
 	}
+	return we;
 }
 
 /**
@@ -770,19 +924,17 @@ static void tag_regex_string(Sentence sent, int i, const char * type)
  * to those of the unknown words
  * so "grok" becomes "grok[?].v" 
  */
-static void handle_unknown_word(Sentence sent, int i, char * s)
+static X_node * handle_unknown_word(Sentence sent, int i, const char * s)
 {
-	char *t;
-	X_node *d;
-	char str[MAX_WORD+1];
+	X_node *d, *we;
 
-	sent->word[i].x = build_word_expressions(sent->dict, UNKNOWN_WORD);
-	if (sent->word[i].x == NULL)
-		assert(FALSE, "UNKNOWN_WORD should have been there");
+	we = build_word_expressions(sent->dict, UNKNOWN_WORD);
+	assert(we, "UNKNOWN_WORD must be defined in the dictionary!");
 
-	for (d = sent->word[i].x; d != NULL; d = d->next)
+	for (d = we; d != NULL; d = d->next)
 	{
-		t = strchr(d->string, '.');
+		char str[MAX_WORD+1];
+		char *t = strchr(d->string, '.');
 		if (t != NULL)
 		{
 			snprintf(str, MAX_WORD, "%.50s[?].%.5s", s, t+1);
@@ -793,16 +945,16 @@ static void handle_unknown_word(Sentence sent, int i, char * s)
 		}
 		d->string = string_set_add(str, sent->string_set);
 	}
+	return we;
 }
 
 /**
  * If a word appears to be mis-spelled, then add alternate
  * spellings. Maybe one of those will do ... 
  */
-static void guess_misspelled_word(Sentence sent, int i, char * s)
+static X_node * guess_misspelled_word(Sentence sent, int i, const char * s)
 {
 	Boolean spelling_ok;
-	char str[MAX_WORD+1];
 	Dictionary dict = sent->dict;
 	X_node *d, *head = NULL;
 	int j, n;
@@ -811,8 +963,7 @@ static void guess_misspelled_word(Sentence sent, int i, char * s)
 	/* Spell-guessing is disabled if no spell-checker is speficified */
 	if (NULL == dict->spell_checker)
 	{
-		handle_unknown_word(sent, i, s);
-		return;
+		return handle_unknown_word(sent, i, s);
 	}
 
 	/* If the spell-checker knows about this word, and we don't ... 
@@ -820,8 +971,7 @@ static void guess_misspelled_word(Sentence sent, int i, char * s)
 	spelling_ok = spellcheck_test(dict->spell_checker, s);
 	if (spelling_ok)
 	{
-		handle_unknown_word(sent, i, s);
-		return;
+		return handle_unknown_word(sent, i, s);
 	}
 
 	/* Else, ask the spell-checker for alternate spellings
@@ -829,23 +979,24 @@ static void guess_misspelled_word(Sentence sent, int i, char * s)
 	n = spellcheck_suggest(dict->spell_checker, &alternates, s);
 	for (j=0; j<n; j++)
 	{
-		if (boolean_reg_dict_lookup(sent->dict, alternates[j]))
+		if (find_word_in_dict(sent->dict, alternates[j]))
 		{
 			X_node *x = build_word_expressions(sent->dict, alternates[j]);
 			head = catenate_X_nodes(x, head);
 		}
 	}
-	sent->word[i].x = head;
 	if (alternates) spellcheck_free_suggest(alternates, n);
 
 	/* Add a [~] to the output to signify that its the result of
 	 * guessing. */
-	for (d = sent->word[i].x; d != NULL; d = d->next)
+	for (d = head; d != NULL; d = d->next)
 	{
+		char str[MAX_WORD+1];
 		const char * t = strchr(d->string, '.');
 		if (t != NULL)
 		{
 			size_t off = t - d->string;
+			off = MIN(off, MAX_WORD-3); /* make room for [~] */
 			strncpy(str, d->string, off);
 			str[off] = 0;
 			strcat(str, "[~]");
@@ -861,8 +1012,9 @@ static void guess_misspelled_word(Sentence sent, int i, char * s)
 	/* If nothing found at all... */
 	if (NULL == head)
 	{
-		handle_unknown_word(sent, i, s);
+		return handle_unknown_word(sent, i, s);
 	}
+	return head;
 }
 
 /**
@@ -892,10 +1044,10 @@ static void guess_misspelled_word(Sentence sent, int i, char * s)
  * word count bythis point.  The viterbi incremental parser won't have
  * this bug.
  */
-int build_sentence_expressions(Sentence sent, Parse_Options opts)
+void build_sentence_expressions(Sentence sent, Parse_Options opts)
 {
 	int i, first_word;  /* the index of the first word after the wall */
-	char *s, temp_word[MAX_WORD+1];
+	const char *s;
 	const char * regex_name;
 	X_node * e;
 	Dictionary dict = sent->dict;
@@ -910,116 +1062,130 @@ int build_sentence_expressions(Sentence sent, Parse_Options opts)
 	 * (nothing special for 1st word) */
 	for (i=0; i<sent->length; i++)
 	{
-		s = sent->word[i].string;
-		if (boolean_dictionary_lookup(sent->dict, s))
+		size_t ialt;
+		for (ialt=0; NULL != sent->word[i].alternatives[ialt]; ialt++)
 		{
-			sent->word[i].x = build_word_expressions(sent->dict, s);
-		}
-		else if ((NULL != (regex_name = match_regex(sent->dict, s))) &&
-		         boolean_dictionary_lookup(sent->dict, regex_name))
-		{
-			tag_regex_string(sent, i, regex_name);
-		}
-		else if (dict->unknown_word_defined && dict->use_unknown_word)
-		{
-			if (opts->use_spell_guess)
+			X_node *we;
+
+			s = sent->word[i].alternatives[ialt];
+			if (boolean_dictionary_lookup(sent->dict, s))
 			{
-				guess_misspelled_word(sent, i, s);
+				we = build_word_expressions(sent->dict, s);
 			}
-			else
+			else if ((NULL != (regex_name = match_regex(sent->dict, s))) &&
+			         boolean_dictionary_lookup(sent->dict, regex_name))
 			{
-				handle_unknown_word(sent, i, s);
+				we = build_regex_expressions(sent, i, regex_name, s);
 			}
-		}
-		else 
-		{
-			/* The reason I can assert this is that the word
-			 * should have been looked up already if we get here.
-			 */
-			assert(FALSE, "I should have found that word.");
-		}
-	}
-
-	/* Under certain cases--if it's the first word of the sentence,
-	 * or if it follows a colon or a quotation mark--a word that's 
-	 * capitalized has to be looked up as an uncapitalized word
-	 * (as well as a capitalized word).
-	 * XXX This rule is English-language-oriented, and should be
-	 * abstracted.
-	 */
-	for (i=0; i<sent->length; i++)
-	{
-		if (! (i == first_word ||
-		      (i > 0 && strcmp(":", sent->word[i-1].string)==0) || 
-		       sent->post_quote[i] == 1)) continue;
-		s = sent->word[i].string;
-
-		/* If the lower-case version of this word is in the dictionary, 
-		 * then add the disjuncts for the lower-case version. The upper
-		 * case version disjuncts had previously come from matching the 
-		 * CAPITALIZED-WORDS regex.
-		 *
-		 * Err .. add the lower-case version only if the lower-case word
-		 * is a common noun or adjective; otherwise, *replace* the
-		 * upper-case word with the lower-case one.  This allows common
-		 * nouns and adjectives to be used for entity names: e.g.
-		 * "Great Southern Union declares bankruptcy", allowing Great
-		 * to be capitalized, while preventing an upper-case "She" being
-		 * used as a proper name in "She declared bankruptcy".
-		 *
-		 * Arghh. This is still messed up. The capitalized-regex runs
-		 * too early, I think. We need to *add* Sue.f (female name Sue)
-		 * even though sue.v (the verb "to sue") is in the dict. So 
-		 * test for capitalized entity names. Glurg. Too much complexity
-		 * here, it seems to me.
-		 *
-		 * This is actually a great example of a combo of an algorithm
-		 * together with a list of words used to determine grammatical
-		 * function.
-		 */
-		if (is_utf8_upper(s))
-		{
-			const char * lc;
-			downcase_utf8_str(temp_word, s, MAX_WORD);
-			lc = string_set_add(temp_word, sent->string_set);
-
-			/* The lower-case dict lookup might trigger regex 
-			 * matches in the dictionary. We want to avoid these.
-			 * e.g. "Cornwallis" triggers both PL-CAPITALIZED_WORDS
-			 * and S-WORDS. Since its not an entity, the regex 
-			 * matches will erroneously discard the upper-case version.
-			 */
-			if (boolean_dictionary_lookup(sent->dict, lc))
+			else if (dict->unknown_word_defined && dict->use_unknown_word)
 			{
-				if (is_entity(sent->dict,s) ||
-				    is_common_entity(sent->dict,lc))
+				if (opts->use_spell_guess)
 				{
-					if (1 < verbosity)
-					{
-						printf ("Info: First word: %s entity=%d common=%d\n", 
-							s, is_entity(sent->dict,s), 
-							is_common_entity(sent->dict,lc));
-					}
-					e = build_word_expressions(sent->dict, lc);
-					sent->word[i].x =
-						catenate_X_nodes(sent->word[i].x, e);
+					we = guess_misspelled_word(sent, i, s);
 				}
 				else
 				{
-					if (1 < verbosity)
-					{
-						printf("Info: First word: %s downcase only\n", lc);
-					}
-					safe_strcpy(s, lc, MAX_WORD);
-					e = build_word_expressions(sent->dict, s);
-					free_X_nodes(sent->word[i].x);
-					sent->word[i].x = e;
+					we = handle_unknown_word(sent, i, s);
 				}
 			}
+			else 
+			{
+				/* The reason I can assert this is that the word
+				 * should have been looked up already if we get here.
+				 */
+				assert(FALSE, "I should have found that word.");
+			}
+
+			/* Under certain cases--if it's the first word of the sentence,
+			 * or if it follows a colon or a quotation mark--a word that's 
+			 * capitalized has to be looked up as an uncapitalized word
+			 * (possibly also as well as a capitalized word).
+			 *
+			 * XXX This rule is English-language-oriented, and should be
+			 * abstracted.
+			 */
+			if ((i == first_word ||
+			     (i > 0 && strcmp(":", sent->word[i-1].alternatives[0])==0) || 
+			     sent->post_quote[i]) &&
+			   (is_utf8_upper(s)))
+			{
+
+				/* If the lower-case version of this word is in the dictionary, 
+				 * then add the disjuncts for the lower-case version. The upper
+				 * case version disjuncts had previously come from matching the 
+				 * CAPITALIZED-WORDS regex.
+				 *
+				 * Err .. add the lower-case version only if the lower-case word
+				 * is a common noun or adjective; otherwise, *replace* the
+				 * upper-case word with the lower-case one.  This allows common
+				 * nouns and adjectives to be used for entity names: e.g.
+				 * "Great Southern Union declares bankruptcy", allowing Great
+				 * to be capitalized, while preventing an upper-case "She" being
+				 * used as a proper name in "She declared bankruptcy".
+				 *
+				 * Arghh. This is still messed up. The capitalized-regex runs
+				 * too early, I think. We need to *add* Sue.f (female name Sue)
+				 * even though sue.v (the verb "to sue") is in the dict. So 
+				 * test for capitalized entity names. Glurg. Too much complexity
+				 * here, it seems to me.
+				 *
+				 * This is actually a great example of a combo of an algorithm
+				 * together with a list of words used to determine grammatical
+				 * function.
+				 */
+				char temp_word[MAX_WORD+1];
+				const char * lc;
+
+				downcase_utf8_str(temp_word, s, MAX_WORD);
+				lc = string_set_add(temp_word, sent->string_set);
+	
+				/* The lower-case dict lookup might trigger regex 
+				 * matches in the dictionary. We want to avoid these.
+				 * e.g. "Cornwallis" triggers both PL-CAPITALIZED_WORDS
+				 * and S-WORDS. Since its not an entity, the regex 
+				 * matches will erroneously discard the upper-case version.
+				 */
+				if (boolean_dictionary_lookup(sent->dict, lc))
+				{
+					if (is_entity(sent->dict,s) ||
+					    is_common_entity(sent->dict,lc))
+					{
+						if (1 < verbosity)
+						{
+							printf ("Info: First word: %s entity=%d common=%d\n", 
+								s, is_entity(sent->dict,s), 
+								is_common_entity(sent->dict,lc));
+						}
+	
+						/* If we are here, then we want both upper and lower case
+						 * expressions. The upper-case ones were build above, so now
+						 * append the lower-case ones. */
+						e = build_word_expressions(sent->dict, lc);
+						we = catenate_X_nodes(we, e);
+					}
+					else
+					{
+						if (1 < verbosity)
+						{
+							printf("Info: First word: %s downcase only\n", lc);
+						}
+	
+						/* If we are here, then we want the lower-case 
+						 * expressions only.  Erase the upper-case ones, built
+						 * previously up above. */
+						sent->word[i].alternatives[ialt] = lc;  /* lc already in sent->string-set */
+						e = build_word_expressions(sent->dict, lc);
+						free_X_nodes(we);
+						we = e;
+					}
+				}
+			}
+
+			/* At last .. concatentate the word expressions we build for
+			 * this alternative. */
+			sent->word[i].x = catenate_X_nodes(sent->word[i].x, we);
 		}
 	}
-
-	return TRUE;
 }
 
 
@@ -1031,27 +1197,32 @@ int build_sentence_expressions(Sentence sent, Parse_Options opts)
  *
  * This code is called only is the 'unkown-words' flag is set.
  */
-int sentence_in_dictionary(Sentence sent)
+Boolean sentence_in_dictionary(Sentence sent)
 {
-	int w, ok_so_far;
-	char * s;
+	Boolean ok_so_far;
+	int w;
+	const char * s;
 	Dictionary dict = sent->dict;
 	char temp[1024];
 
 	ok_so_far = TRUE;
 	for (w=0; w<sent->length; w++)
 	{
-		s = sent->word[w].string;
-		if (!boolean_reg_dict_lookup(dict, s))
+		size_t ialt;
+		for (ialt=0; NULL != sent->word[w].alternatives[ialt]; ialt++)
 		{
-			if (ok_so_far)
+			s = sent->word[w].alternatives[ialt];
+			if (!find_word_in_dict(dict, s))
 			{
-				safe_strcpy(temp, "The following words are not in the dictionary:", sizeof(temp));
-				ok_so_far = FALSE;
+				if (ok_so_far)
+				{
+					safe_strcpy(temp, "The following words are not in the dictionary:", sizeof(temp));
+					ok_so_far = FALSE;
+				}
+				safe_strcat(temp, " \"", sizeof(temp));
+				safe_strcat(temp, s, sizeof(temp));
+				safe_strcat(temp, "\"", sizeof(temp));
 			}
-			safe_strcat(temp, " \"", sizeof(temp));
-			safe_strcat(temp, sent->word[w].string, sizeof(temp));
-			safe_strcat(temp, "\"", sizeof(temp));
 		}
 	}
 	if (!ok_so_far)
