@@ -383,7 +383,7 @@ int match_in_connector_set(Sentence sent, Connector_set *conset, Connector * c,
  * connector-matching function used in parsing - except it ignores
  * "priority" (used to handle fat links)
  */
-static int easy_match(const char * s, const char * t)
+static Boolean easy_match(const char * s, const char * t)
 {
 	while(isupper((int)*s) || isupper((int)*t)) {
 		if (*s != *t) return FALSE;
@@ -405,33 +405,57 @@ static int easy_match(const char * s, const char * t)
  * word_has_connector() -- return TRUE if dictionary expression has connector
  * This function takes a dict_node (corresponding to an entry in a
  * given dictionary), a string (representing a connector), and a
- * direction (0 = right-pointing, 1 = left-pointing); it returns 1
+ * direction (+ = right-pointing, '-' = left-pointing); it returns true
  * if the dictionary expression for the word includes the connector,
- * 0 otherwise.  This can be used to see if a word is in a certain
+ * false otherwise.  This can be used to see if a word is in a certain
  * category (checking for a category connector in a table), or to see
  * if a word has a connector in a normal dictionary. The connector
  * check uses a "smart-match", the same kind used by the parser.
  */
-int word_has_connector(Dict_node * dn, const char * cs, int direction)
+#if CRAZY_OBESE_CHECKING_AGLO
+Boolean word_has_connector(Dict_node * dn, const char * cs, char direction)
 {
-	Connector * c2=NULL;
-	Disjunct * d, *d0;
-	if(dn == NULL) return -1;
+	Connector * c2 = NULL;
+	Disjunct *d, *d0;
+	if (dn == NULL) return FALSE;
 	d0 = d = build_disjuncts_for_dict_node(dn);
-	if(d == NULL) return 0;
-	for(; d!=NULL; d=d->next) {
-		if(direction==0) c2 = d->right;
-		if(direction==1) c2 = d->left;
-		for(; c2!=NULL; c2=c2->next) {
-			if(easy_match(c2->string, cs)==1) {
+	if (d == NULL) return FALSE;
+	for (; d != NULL; d = d->next) {
+		if (direction == '+') c2 = d->right;
+		if (direction == '-') c2 = d->left;
+		for (; c2 != NULL; c2 = c2->next) {
+			if (easy_match(c2->string, cs)) {
 				free_disjuncts(d0);
-				return 1;
+				return TRUE;
 			}
 		}
 	}
 	free_disjuncts(d0);
-	return 0;
+	return FALSE;
 }
+#else /* CRAZY_OBESE_CHECKING_AGLO */
+
+static Boolean exp_has_connector(Exp * e, const char * cs, char direction)
+{
+	E_list * el;
+	if (e->type == CONNECTOR_type)
+	{
+		if (direction == e->dir)
+			return easy_match(e->u.string, cs);
+		return FALSE;
+	}
+	for (el = e->u.l; el != NULL; el = el->next)
+	{
+		if (exp_has_connector(el->e, cs, direction)) return TRUE;
+	}
+	return FALSE;
+}
+
+Boolean word_has_connector(Dict_node * dn, const char * cs, char direction)
+{
+	return exp_has_connector(dn->exp, cs, direction);
+}
+#endif /* CRAZY_OBESE_CHECKING_AGLO */
 
 /* ======================================================== */
 /* Dictionary utilities ... */
@@ -483,16 +507,6 @@ int word_contains(Dictionary dict, const char * word, const char * macro)
 	return ret;
 }
 
-Dict_node * list_whole_dictionary(Dict_node *root, Dict_node *dn)
-{
-	Dict_node *c, *d;
-	if (root == NULL) return dn;
-	c = (Dict_node *) xalloc(sizeof(Dict_node));
-	*c = *root;
-	d = list_whole_dictionary(root->left, dn);
-	c->right = list_whole_dictionary(root->right, d);
-	return c;
-}
 
 #define PAST_TENSE_FORM_MARKER "<marker-past>"
 #define ENTITY_MARKER          "<marker-entity>"
