@@ -47,12 +47,11 @@ void        exfree_link(Link *);
 Connector_set * connector_set_create(Exp *e);
 void connector_set_delete(Connector_set * conset);
 int match_in_connector_set(Sentence, Connector_set *conset, Connector * c, int d);
-int word_has_connector(Dict_node *, const char *, int);
+Boolean word_has_connector(Dict_node *, const char *, char);
 
 
 /* Dictionary utilities ... */
 int word_contains(Dictionary dict, const char * word, const char * macro);
-Dict_node * list_whole_dictionary(Dict_node *, Dict_node *);
 
 static inline int string_hash(const char *s)
 {
