@@ -36,24 +36,15 @@
 
 #include <errno.h>
 #include <locale.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <wchar.h>
 
 /* Used for terminal resizing */
 #ifndef _WIN32
-#include <langinfo.h>
 #include <termios.h>
 #include <sys/ioctl.h>
 #include <fcntl.h>
 #include <unistd.h>
 #endif
 
-#ifdef HAVE_EDITLINE
-#include <editline/readline.h>
-#endif 
-
 #ifdef _MSC_VER
 #define LINK_GRAMMAR_DLL_EXPORT 0
 #endif
@@ -61,23 +52,10 @@
 #include "link-includes.h"
 #include "command-line.h"
 #include "expand.h"
+#include "lg_readline.h"
+#include "utilities.h"     /* For MSVC portability */
 #include "../viterbi/viterbi.h"
 
-#ifndef FALSE
-#define FALSE 0
-#endif
-
-#ifndef TRUE
-#define TRUE 1
-#endif
-
-#if !defined(MIN)
-#define MIN(X,Y)  ( ((X) < (Y)) ? (X) : (Y))
-#endif
-#if !defined(MAX)
-#define MAX(X,Y)  ( ((X) > (Y)) ? (X) : (Y))
-#endif
-
 #define MAX_INPUT 1024
 #define DISPLAY_MAX 1024
 #define COMMENT_CHAR '%'  /* input lines beginning with this are ignored */
@@ -102,6 +80,12 @@ fget_input_string(FILE *in, FILE *out, Parse_Options opts)
 	static char * pline = NULL;
 	const char * prompt = "linkparser> ";
 
+	if (NULL == in)
+	{
+		if (pline) free(pline);
+		return NULL;
+	}
+
 	if (in != stdin)
 	{
 		static char input_string[MAX_INPUT];
@@ -123,18 +107,15 @@ fget_input_string(FILE *in, FILE *out, Parse_Options opts)
 	}
 	input_pending = FALSE;
 	if (pline) free(pline);
-	pline = readline(prompt);
+	pline = lg_readline(prompt);
 
-	/* Save non-blank lines */
-	if (pline && *pline)
-	{
-		if (*pline) add_history(pline);
-	}
 	return pline;
 
 #else
 	static char input_string[MAX_INPUT];
 
+	if (NULL == in) return NULL; 
+
 	if ((!parse_options_get_batch_mode(opts)) && 
 	    (verbosity > 0) && 
 	    (!input_pending))
@@ -683,12 +664,12 @@ int main(int argc, char * argv[])
 	verbosity = parse_options_get_verbosity(opts);
 	check_winsize(opts);
 #ifdef _WIN32
-	parse_options_set_screen_width(popts, 79);
+	parse_options_set_screen_width(opts, 79);
 #endif
 
-	prt_error("Info: Dictionary version %s.\n",
+	prt_error("Info: Dictionary version %s.",
 		linkgrammar_get_dict_version(dict));
-	prt_error("Info: Library version %s. Enter \"!help\" for help.\n",
+	prt_error("Info: Library version %s. Enter \"!help\" for help.",
 		linkgrammar_get_version());
 
 	/* Main input loop */
@@ -883,6 +864,7 @@ int main(int argc, char * argv[])
 	parse_options_delete(panic_parse_opts);
 	parse_options_delete(opts);
 	dictionary_delete(dict);
+	fget_input_string(NULL, NULL, NULL);
 
 	printf ("Bye.\n");
 	return 0;
