@@ -42,6 +42,7 @@ const string type_name(AtomType t)
 
 bool Atom::operator==(const Atom* other) const
 {
+	if (!other) return false;
 	if (other == this) return true;
 	if (other->_type == this->_type) return true;
 	return false;
@@ -49,6 +50,7 @@ bool Atom::operator==(const Atom* other) const
 
 bool Node::operator==(const Atom* other) const
 {
+	if (!other) return false;
 	if (other == this) return true;
 	if (other->get_type() != this->get_type()) return false;
 	const Node* nother = dynamic_cast<const Node*>(other);
@@ -58,6 +60,7 @@ bool Node::operator==(const Atom* other) const
 
 bool Link::operator==(const Atom* other) const
 {
+	if (!other) return false;
 	if (other == this) return true;
 	if (other->get_type() != this->get_type()) return false;
 	const Link* lother = dynamic_cast<const Link*>(other);
@@ -72,7 +75,7 @@ bool Link::operator==(const Atom* other) const
 
 std::ostream& do_prt(std::ostream& out, const Atom* a, int ilvl)
 {
-	const char *indent_str = "  ";
+	static const char *indent_str = "  ";
 	const Node* n = dynamic_cast<const Node*>(a);
 	if (n)
 	{
@@ -88,7 +91,8 @@ std::ostream& do_prt(std::ostream& out, const Atom* a, int ilvl)
 		out << type_name(l->get_type()) <<" :" << endl;
 
 		ilvl++;
-		for (int i=0; i < l->get_arity(); i++)
+		size_t lsz = l->get_arity();
+		for (int i=0; i < lsz; i++)
 		{
 			do_prt(out, l->get_outgoing_atom(i), ilvl);
 		}
