@@ -15,6 +15,9 @@
 #include <iostream>
 #include <string>
 #include <vector>
+#include <gc/gc_allocator.h>
+
+#include "garbage.h"
 
 namespace link_grammar {
 namespace viterbi {
@@ -33,7 +36,8 @@ class TV
 enum AtomType
 {
 	// Node types
-	WORD = 1,
+	NODE = 1,
+	WORD,       // a word
 	LING_TYPE,  // a pair of merged connectors (LG LINK TYPE)
 	// META,       // special-word, e.g. LEFT-WALL, RIGHT-WALL
 	CONNECTOR,  // e.g. S+ 
@@ -47,6 +51,8 @@ enum AtomType
 	WORD_DISJ,  // word, followed by a single disjunct for that word.
 	LING,       // two connected connectors, (LGLINK) e.g. Dmcn w/o direction info
 	STATE_PAIR, // Current state and corresponding output.
+
+	RULE,       // Base class for graph re-write rules
 };
 
 #define OPTIONAL_CLAUSE "0"
@@ -54,11 +60,17 @@ enum AtomType
 /* Base class for Nodes and Links */
 /* Atoms are not mutable, except for the TV value. That is, you cannot
  * change the type of the atom.  In particular, all methods are const.
+ *
+ * All atoms are automatically garbage-collected.
  */
-class Atom
+class Atom : public gc
 {
 	public:
-		Atom(AtomType type) : _type(type) {}
+		Atom(AtomType type) : _type(type)
+		{
+			// Marking stubborn, since its immutable.
+			GC_change_stubborn(this);
+		}
 		AtomType get_type() const { return _type; }
 		TV tv;
 		virtual bool operator==(const Atom*) const;
@@ -78,6 +90,9 @@ class Atom
 class Node : public Atom
 {
 	public:
+		Node(const std::string& n)
+			: Atom(NODE), _name(n) {}
+
 		Node(AtomType t, const std::string& n)
 			: Atom(t), _name(n) {}
 
@@ -88,8 +103,9 @@ class Node : public Atom
 		const std::string _name;
 };
 
-// Uhhhh ... 
-typedef std::vector<Atom*> OutList;
+/// All outgoing lists will be handled as vectors.
+// Must use the bdw-gc allocator to track these pointers.
+typedef std::vector<Atom*, gc_allocator<Atom*> > OutList;
 
 /*
  * Links hold a bunch of atoms
@@ -115,6 +131,10 @@ class Link : public Atom
 			: Atom(t), _oset(({OutList o(1,a); o.push_back(b);
 			                   o.push_back(c); o;}))
 		{}
+		Link(AtomType t, Atom* a, Atom* b, Atom* c, Atom* d)
+			: Atom(t), _oset(({OutList o(1,a); o.push_back(b);
+			                   o.push_back(c); o.push_back(d); o;}))
+		{}
 		size_t get_arity() const { return _oset.size(); }
 		Atom* get_outgoing_atom(size_t pos) const { return _oset.at(pos); }
 		const OutList& get_outgoing_set() const { return _oset; }
