@@ -1,5 +1,5 @@
 /*************************************************************************/
-/* Copyright (c) 2012 Linas Vepstas <linasvepstas@gmail.com>             */
+/* Copyright (c) 2012, 2013 Linas Vepstas <linasvepstas@gmail.com>       */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the Viterbi parsing system is subject to the terms of the      */
@@ -87,9 +87,16 @@ class Set : public Link
 		Set(Atom* singleton)
 			: Link(SET, OutList(1, singleton))
 		{}
+		Set()
+			: Link(SET)
+		{}
 
 		// See the C file for documentation
 		Set* flatten() const { return new Set(flatset()); }
+		Atom* super_flatten() const;
+
+		// Add (append) other set  to this set.
+		Set* add(const Set*);
 
 protected:
 		/// The sole purpose of this ctor is to allow inheritance.
@@ -117,6 +124,9 @@ class Seq : public Set
 		Seq(Atom* singleton)
 			: Set(SEQ, OutList(1, singleton))
 		{}
+		Seq(Atom* a, Atom* b)
+			: Set(SEQ, ({OutList o(1,a); o.push_back(b); o;}))
+		{}
 
 		// See the Set class for documentation
 		Seq* flatten() const { return new Seq(flatset()); }
@@ -131,6 +141,36 @@ protected:
 		{}
 };
 
+/// Unordered OR of all children
+class Or : public Set
+{
+	public:
+		Or()
+			: Set(OR)
+		{}
+		Or(const OutList& ol)
+			: Set(OR, ol)
+		{}
+		Or(Atom* singleton)
+			: Set(OR, OutList(1, singleton))
+		{}
+		Or(Atom* a, Atom* b)
+			: Set(OR, ({OutList o(1,a); o.push_back(b); o;}))
+		{}
+		Or(Atom* a, Atom* b, Atom* c)
+			: Set(OR, ({OutList o(1,a); o.push_back(b); o.push_back(c); o;}))
+		{}
+
+		// Return disjunctive normal form (DNF)
+		Or* disjoin() const;
+
+		// See the Set class for documentation
+		Or* flatten() const { return new Or(flatset()); }
+
+		// Remove repeated entries
+		Or* uniq() const;
+};
+
 /// Ordered sequence
 /// And inherits from Seq, since the order of the atoms in
 /// its outgoing set is important.
@@ -143,6 +183,20 @@ class And : public Seq
 		And(const OutList& ol)
 			: Seq(AND, ol)
 		{}
+		And(Atom* singleton)
+			: Seq(AND, OutList(1, singleton))
+		{}
+		And(Atom* a, Atom* b)
+			: Seq(AND, ({OutList o(1,a); o.push_back(b); o;}))
+		{}
+		And(Atom* a, Atom* b, Atom* c)
+			: Seq(AND, ({OutList o(1,a); o.push_back(b); o.push_back(c); o;}))
+		{}
+
+
+		// Return disjunctive normal form (DNF)
+		// Does not modify this atom; just returns a new one.
+		Or* disjoin();
 
 		// See the Set class for documentation
 		And* flatten() const { return new And(flatset()); }
@@ -151,21 +205,6 @@ class And : public Seq
 		And* clean() const;
 };
 
-/// Unordered OR of all children
-class Or : public Set
-{
-	public:
-		Or()
-			: Set(OR)
-		{}
-		Or(const OutList& ol)
-			: Set(OR, ol)
-		{}
-
-		// See the Set class for documentation
-		Or* flatten() const { return new Or(flatset()); }
-};
-
 /// Create a ling-grammar link. This will be of the form:
 ///     LING:
 ///        Ling_TYPE "MVa"
@@ -232,6 +271,8 @@ class WordCset : public Link
 		{
 			return _oset[1];
 		}
+		bool has_left_pointers() const;
+		WordCset* flatten();
 };
 
 /// A pair of two sequences.  The first sequence is the state, the
@@ -240,7 +281,8 @@ class WordCset : public Link
 ///
 ///    STATE_PAIR :
 ///       SEQ
-///           ...
+///           WORD_CSET ...
+///           WORD_CSET ...
 ///       SEQ
 ///           ...
 ///
