@@ -1,5 +1,5 @@
 /*************************************************************************/
-/* Copyright (c) 2012 Linas Vepstas <linasvepstas@gmail.com>             */
+/* Copyright (c) 2012, 2013 Linas Vepstas <linasvepstas@gmail.com>       */
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the Viterbi parsing system is subject to the terms of the      */
@@ -27,6 +27,7 @@
 #include "connect.h"
 #include "connector-utils.h"
 #include "disjoin.h"
+#include "garbage.h"
 #include "parser.h"
 #include "state.h"
 #include "viterbi.h"
@@ -42,6 +43,7 @@ Parser::Parser(Dictionary dict)
 	: _dict(dict), _alternatives(NULL)
 {
 	DBG(cout << "=============== Parser ctor ===============" << endl);
+	lg_init_gc();
 	initialize_state();
 }
 
@@ -79,7 +81,7 @@ Atom * Parser::lg_exp_to_atom(Exp* exp)
 	if (NULL == el)
 		return new Connector(OPTIONAL_CLAUSE);
 
-	// The data structure that link-grammar uses for connector
+	// The C data structure that link-grammar uses for connector
 	// expressions is totally insane, as witnessed by the loop below.
 	// Anyway: operators are infixed, i.e. are always binary,
 	// with exp->u.l->e being the left hand side, and
@@ -113,8 +115,8 @@ Atom * Parser::lg_exp_to_atom(Exp* exp)
  * Return atomic formula connector expression for the given word.
  *
  * This looks up the word in the link-grammar dictionary, and converts
- * the resulting link-grammar connective expression into an atomic
- * formula.
+ * the resulting link-grammar connective expression into an formula
+ * composed of atoms.
  */
 Set * Parser::word_consets(const string& word)
 {
@@ -126,15 +128,13 @@ Set * Parser::word_consets(const string& word)
 	for (Dict_node*dn = dn_head; dn; dn= dn->right)
 	{
 		Exp* exp = dn->exp;
-cout<<"duude the word: " << dn->string << ": ";
-print_expression(exp);
+		DBG({cout << "=============== Parser word: " << dn->string << ": ";
+			print_expression(exp); });
 
 		Atom *dj = lg_exp_to_atom(exp);
-cout <<"duuude the bbbbbbbbbbbbbbbbb efore it is "<<dj<<endl;
 		dj = disjoin(dj);
-cout <<"duuude the ddddddddddddddddddddddddddddddddddddddsijoined thing is "<<dj<<endl;
 
-		// First atom at the from of the outgoing set is the word itself.
+		// First atom at the front of the outgoing set is the word itself.
 		// Second atom is the first disjuct that must be fulfilled.
 		Word* nword = new Word(dn->string);
 		djset.push_back(new WordCset(nword, dj));
@@ -152,15 +152,15 @@ void Parser::initialize_state()
 
 	Set *wall_disj = word_consets(wall_word);
 
-	// We are expecting the initiall wall to be unique.
+	// We are expecting the initial wall to be unique.
 	assert(wall_disj->get_arity() == 1, "Unexpected wall structure");
 	OutList state_vec;
-	Atom* word_cset = wall_disj->get_outgoing_atom(0);
+	Atom* wall_cset = wall_disj->get_outgoing_atom(0);
 
 	// Initial state: no output, and the wall cset.
 	_alternatives = new Set(
 		new StatePair(
-			new Seq(word_cset),
+			new Seq(wall_cset),
 			new Seq()
 		)
 	);
@@ -179,16 +179,16 @@ void Parser::stream_word(const string& word)
 		return;
 	}
 
-assert(1 == djset->get_arity(), "Multiple dict entries not handled");
-
 	// Try to add each dictionary entry to the parse state.
+	Set* new_alts = new Set();
 	for (int i = 0; i < djset->get_arity(); i++)
 	{
+		Atom* cset = djset->get_outgoing_atom(i);
 		State stset(_alternatives);
-		stset.stream_word_conset(dynamic_cast<WordCset*>(djset->get_outgoing_atom(i)));
-// XXX this can't possibly be right ...
-_alternatives = stset.get_alternatives();
+		stset.stream_word_conset(dynamic_cast<WordCset*>(cset));
+		new_alts = new_alts->add(stset.get_alternatives());
 	}
+	_alternatives = new_alts;
 }
 
 // ===================================================================
@@ -232,6 +232,23 @@ void viterbi_parse(Dictionary dict, const char * sentence)
 	Parser pars(dict);
 
 	pars.streamin(sentence);
+	Link* alts = pars.get_alternatives();
+
+	/* Print some diagnostic outputs ... for now. Remove whhen finished. */
+	size_t num_alts = alts->get_arity();
+	printf("Found %lu alternatives\n", num_alts);
+	for (size_t i=0; i<num_alts; i++)
+	{
+		Atom* a = alts->get_outgoing_atom(i);
+		StatePair* sp = dynamic_cast<StatePair*>(a);
+		Seq* state = sp->get_state();
+		size_t state_sz = state->get_arity();
+		if (0 == state_sz)
+		{
+			cout << "\nAlternative =============== " << i << endl;
+			cout << sp->get_output() << endl;
+		}
+	}
 }
 
 } // namespace viterbi
